# --------------------------------------------------------------------

import re
import os
import sys
import glob
import copy
import warnings
from distutils import log
from distutils import sysconfig
from distutils.util import execute
from distutils.util import split_quoted
from distutils.errors import DistutilsError
from distutils.text_file import TextFile


try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

try:
    import setuptools
except ImportError:
    setuptools = None

if setuptools:
    from setuptools import setup as _setup
    from setuptools import Extension as _Extension
    from setuptools import Command
else:
    from distutils.core import setup as _setup
    from distutils.core import Extension as _Extension
    from distutils.core import Command


def import_command(cmd):
    try:
        from importlib import import_module
    except ImportError:

        def import_module(n):
            return __import__(n, fromlist=[None])

    try:
        if not setuptools:
            raise ImportError
        mod = import_module('setuptools.command.' + cmd)
        return getattr(mod, cmd)
    except ImportError:
        mod = import_module('distutils.command.' + cmd)
        return getattr(mod, cmd)


_config = import_command('config')
_build = import_command('build')
_build_ext = import_command('build_ext')
_install = import_command('install')

try:
    from setuptools import modified
except ImportError:
    try:
        from setuptools import dep_util as modified
    except ImportError:
        from distutils import dep_util as modified

try:
    from packaging.version import Version
except ImportError:
    try:
        from setuptools.extern.packaging.version import Version
    except ImportError:
        from distutils.version import StrictVersion as Version

# --------------------------------------------------------------------

# Cython

CYTHON = '3.0.0'


def cython_req():
    return CYTHON


def cython_chk(VERSION, verbose=True):
    #
    def warn(message):
        if not verbose:
            return
        ruler, ws, nl = '*' * 80, ' ', '\n'
        pyexe = sys.executable
        advise = '$ %s -m pip install --upgrade cython' % pyexe

        def printer(*s):
            sys.stderr.write(' '.join(s) + '\n')

        printer(ruler, nl)
        printer(ws, message, nl)
        printer(ws, ws, advise, nl)
        printer(ruler)

    #
    try:
        import Cython
    except ImportError:
        warn('You need Cython to generate C source files.')
        return False
    #
    CYTHON_VERSION = Cython.__version__
    m = re.match(r'(\d+\.\d+(?:\.\d+)?).*', CYTHON_VERSION)
    if not m:
        warn(f'Cannot parse Cython version string {CYTHON_VERSION!r}')
        return False
    REQUIRED = Version(VERSION)
    PROVIDED = Version(m.groups()[0])
    if PROVIDED < REQUIRED:
        warn(f'You need Cython >= {VERSION} (you have version {CYTHON_VERSION})')
        return False
    #
    if verbose:
        log.info('using Cython %s' % CYTHON_VERSION)
    return True


def cython_run(
    source,
    target=None,
    depends=(),
    includes=(),
    workdir=None,
    force=False,
    VERSION='0.0',
):
    if target is None:
        target = os.path.splitext(source)[0] + '.c'
    cwd = os.getcwd()
    try:
        if workdir:
            os.chdir(workdir)
        alldeps = [source]
        for dep in depends:
            alldeps += glob.glob(dep)
        if not (force or modified.newer_group(alldeps, target)):
            log.debug("skipping '%s' -> '%s' (up-to-date)", source, target)
            return
    finally:
        os.chdir(cwd)
    require = 'Cython >= %s' % VERSION
    if setuptools and not cython_chk(VERSION, verbose=False):
        if sys.modules.get('Cython'):
            removed = getattr(sys.modules['Cython'], '__version__', '')
            log.info('removing Cython %s from sys.modules' % removed)
            pkgname = re.compile(r'cython(\.|$)', re.IGNORECASE)
            for modname in list(sys.modules.keys()):
                if pkgname.match(modname):
                    del sys.modules[modname]
        try:
            install_setup_requires = setuptools._install_setup_requires
            with warnings.catch_warnings():
                if hasattr(setuptools, 'SetuptoolsDeprecationWarning'):
                    category = setuptools.SetuptoolsDeprecationWarning
                    warnings.simplefilter('ignore', category)
                log.info("fetching build requirement '%s'" % require)
                install_setup_requires({'setup_requires': [require]})
        except Exception:
            log.info("failed to fetch build requirement '%s'" % require)
    if not cython_chk(VERSION):
        raise DistutilsError("unsatisfied build requirement '%s'" % require)
    #
    log.info("cythonizing '%s' -> '%s'", source, target)
    from cythonize import cythonize

    args = []
    if workdir:
        args += ['--working', workdir]
    args += [source]
    if target:
        args += ['--output-file', target]
    err = cythonize(args)
    if err:
        raise DistutilsError(f"Cython failure: '{source}' -> '{target}'")


# --------------------------------------------------------------------


def fix_config_vars(names, values):
    values = list(values)
    if 'CONDA_BUILD' in os.environ:
        return values
    if sys.platform == 'darwin':
        if 'ARCHFLAGS' in os.environ:
            ARCHFLAGS = os.environ['ARCHFLAGS']
            for i, flag in enumerate(list(values)):
                flag, count = re.subn(r'-arch\s+\w+', ' ', str(flag))
                if count and ARCHFLAGS:
                    flag = flag + ' ' + ARCHFLAGS
                values[i] = flag
        if 'SDKROOT' in os.environ:
            SDKROOT = os.environ['SDKROOT']
            for i, flag in enumerate(list(values)):
                flag, count = re.subn(r'-isysroot [^ \t]*', ' ', str(flag))
                if count and SDKROOT:
                    flag = flag + ' ' + '-isysroot ' + SDKROOT
                values[i] = flag
    return values


def get_config_vars(*names):
    # Core Python configuration
    values = sysconfig.get_config_vars(*names)
    # Do any distutils flags fixup right now
    return fix_config_vars(names, values)


# --------------------------------------------------------------------


class PetscConfig:
    def __init__(self, petsc_dir, petsc_arch, dest_dir=None):
        if dest_dir is None:
            dest_dir = os.environ.get('DESTDIR')
        self.configdict = {}
        if not petsc_dir:
            raise DistutilsError('PETSc not found')
        if not os.path.isdir(petsc_dir):
            raise DistutilsError('invalid PETSC_DIR: %s' % petsc_dir)
        self.version = self._get_petsc_version(petsc_dir)
        self.configdict = self._get_petsc_config(petsc_dir, petsc_arch)
        self.PETSC_DIR = self['PETSC_DIR']
        self.PETSC_ARCH = self['PETSC_ARCH']
        self.DESTDIR = dest_dir
        language_map = {'CONLY': 'c', 'CXXONLY': 'c++'}
        self.language = language_map[self['PETSC_LANGUAGE']]

    def __getitem__(self, item):
        return self.configdict[item]

    def get(self, item, default=None):
        return self.configdict.get(item, default)

    def configure(self, extension, compiler=None):
        self.configure_extension(extension)
        if compiler is not None:
            self.configure_compiler(compiler)

    def _get_petsc_version(self, petsc_dir):
        import re

        version_re = {
            'major': re.compile(r'#define\s+PETSC_VERSION_MAJOR\s+(\d+)'),
            'minor': re.compile(r'#define\s+PETSC_VERSION_MINOR\s+(\d+)'),
            'micro': re.compile(r'#define\s+PETSC_VERSION_SUBMINOR\s+(\d+)'),
            'release': re.compile(r'#define\s+PETSC_VERSION_RELEASE\s+(-*\d+)'),
        }
        petscversion_h = os.path.join(petsc_dir, 'include', 'petscversion.h')
        with open(petscversion_h, 'rt') as f:
            data = f.read()
        major = int(version_re['major'].search(data).groups()[0])
        minor = int(version_re['minor'].search(data).groups()[0])
        micro = int(version_re['micro'].search(data).groups()[0])
        release = int(version_re['release'].search(data).groups()[0])
        return (major, minor, micro), (release == 1)

    def _get_petsc_config(self, petsc_dir, petsc_arch):
        from os.path import join, isdir, exists

        PETSC_DIR = petsc_dir
        PETSC_ARCH = petsc_arch
        #
        confdir = join('lib', 'petsc', 'conf')
        if not (PETSC_ARCH and isdir(join(PETSC_DIR, PETSC_ARCH))):
            petscvars = join(PETSC_DIR, confdir, 'petscvariables')
            PETSC_ARCH = makefile(open(petscvars, 'rt')).get('PETSC_ARCH')
        if not (PETSC_ARCH and isdir(join(PETSC_DIR, PETSC_ARCH))):
            PETSC_ARCH = ''
        #
        variables = join(PETSC_DIR, confdir, 'variables')
        if not exists(variables):
            variables = join(PETSC_DIR, PETSC_ARCH, confdir, 'variables')
        petscvariables = join(PETSC_DIR, PETSC_ARCH, confdir, 'petscvariables')
        #
        with open(variables) as f:
            contents = f.read()
        with open(petscvariables) as f:
            contents += f.read()
        #
        confstr = 'PETSC_DIR  = %s\n' % PETSC_DIR
        confstr += 'PETSC_ARCH = %s\n' % PETSC_ARCH
        confstr += contents
        return makefile(StringIO(confstr))

    def _configure_ext(self, ext, dct, append=False):
        extdict = ext.__dict__
        for key, values in dct.items():
            if key in extdict:
                for value in values:
                    if value not in extdict[key]:
                        if not append:
                            extdict[key].insert(0, value)
                        else:
                            extdict[key].append(value)

    def configure_extension(self, extension):
        # includes and libraries
        # paths in PETSc config files point to final installation location, but
        # we might be building against PETSc in staging location (DESTDIR) when
        # DESTDIR is set, so append DESTDIR (if nonempty) to those paths
        petsc_inc = flaglist(prepend_to_flags(self.DESTDIR, self['PETSC_CC_INCLUDES']))
        lib_flags = prepend_to_flags(
            self.DESTDIR,
            '-L{} {}'.format(self['PETSC_LIB_DIR'], self['PETSC_LIB_BASIC']),
        )
        petsc_lib = flaglist(lib_flags)
        # runtime_library_dirs is not supported on Windows
        if sys.platform != 'win32':
            # if DESTDIR is set, then we're building against PETSc in a staging
            # directory, but rpath needs to point to final install directory.
            rpath = [strip_prefix(self.DESTDIR, self['PETSC_LIB_DIR'])]
            if sys.modules.get('petsc') is not None:
                if sys.platform == 'darwin':
                    rpath = ['@loader_path/../../petsc/lib']
                else:
                    rpath = ['$ORIGIN/../../petsc/lib']
            petsc_lib['runtime_library_dirs'].extend(rpath)
        # Link in extra libraries on static builds
        if self['BUILDSHAREDLIB'] != 'yes':
            petsc_ext_lib = split_quoted(self['PETSC_EXTERNAL_LIB_BASIC'])
            petsc_lib['extra_link_args'].extend(petsc_ext_lib)
        self._configure_ext(extension, petsc_inc, append=True)
        self._configure_ext(extension, petsc_lib)

    def configure_compiler(self, compiler):
        if compiler.compiler_type != 'unix':
            return
        getenv = os.environ.get
        # distutils C/C++ compiler
        (cc, cflags, ccshared, cxx) = get_config_vars('CC', 'CFLAGS', 'CCSHARED', 'CXX')
        ccshared = getenv('CCSHARED', ccshared or '')
        cflags = getenv('CFLAGS', cflags or '')
        cflags = cflags.replace('-Wstrict-prototypes', '')
        # distutils linker
        (ldflags, ldshared, so_ext) = get_config_vars('LDFLAGS', 'LDSHARED', 'SO')
        ld = cc
        ldshared = getenv('LDSHARED', ldshared)
        ldflags = getenv('LDFLAGS', cflags + ' ' + (ldflags or ''))
        ldcmd = split_quoted(ld) + split_quoted(ldflags)
        ldshared = [
            flg
            for flg in split_quoted(ldshared)
            if flg not in ldcmd and (flg.find('/lib/spack/env') < 0) and (flg.find('/libexec/spack/') < 0)
        ]
        ldshared = str.join(' ', ldshared)

        #
        def get_flags(cmd):
            if not cmd:
                return ''
            cmd = split_quoted(cmd)
            if os.path.basename(cmd[0]) == 'xcrun':
                del cmd[0]
                while True:
                    if cmd[0] == '-sdk':
                        del cmd[0:2]
                        continue
                    if cmd[0] == '-log':
                        del cmd[0]
                        continue
                    break
            return ' '.join(cmd[1:])

        # PETSc C compiler
        PCC = self['PCC']
        PCC_FLAGS = get_flags(cc) + ' ' + self['PCC_FLAGS']
        PCC_FLAGS = PCC_FLAGS.replace('-fvisibility=hidden', '')
        PCC_FLAGS = PCC_FLAGS.replace('-Wpedantic', '-Wno-pedantic')
        PCC_FLAGS = PCC_FLAGS.replace('-Wextra-semi-stmt', '-Wno-extra-semi-stmt')
        PCC = getenv('PCC', PCC) + ' ' + getenv('PCCFLAGS', PCC_FLAGS)
        PCC_SHARED = str.join(' ', (PCC, ccshared, cflags))
        # PETSc C++ compiler
        PCXX = PCC if self.language == 'c++' else self.get('CXX', cxx)
        # PETSc linker
        PLD = self['PCC_LINKER']
        PLD_FLAGS = get_flags(ld) + ' ' + self['PCC_LINKER_FLAGS']
        PLD_FLAGS = PLD_FLAGS.replace('-fvisibility=hidden', '')
        PLD = getenv('PLD', PLD) + ' ' + getenv('PLDFLAGS', PLD_FLAGS)
        PLD_SHARED = str.join(' ', (PLD, ldshared, ldflags))
        #
        compiler.set_executables(
            compiler=PCC,
            compiler_cxx=PCXX,
            linker_exe=PLD,
            compiler_so=PCC_SHARED,
            linker_so=PLD_SHARED,
        )
        compiler.shared_lib_extension = so_ext

    def log_info(self):
        PETSC_DIR = self['PETSC_DIR']
        PETSC_ARCH = self['PETSC_ARCH']
        version = '.'.join([str(i) for i in self.version[0]])
        release = ('development', 'release')[self.version[1]]
        version_info = version + ' ' + release
        integer_size = '%s-bit' % self['PETSC_INDEX_SIZE']
        scalar_type = self['PETSC_SCALAR']
        precision = self['PETSC_PRECISION']
        language = self['PETSC_LANGUAGE']
        compiler = self['PCC']
        linker = self['PCC_LINKER']
        log.info('PETSC_DIR:    %s' % PETSC_DIR)
        log.info('PETSC_ARCH:   %s' % PETSC_ARCH)
        log.info('version:      %s' % version_info)
        log.info('integer-size: %s' % integer_size)
        log.info('scalar-type:  %s' % scalar_type)
        log.info('precision:    %s' % precision)
        log.info('language:     %s' % language)
        log.info('compiler:     %s' % compiler)
        log.info('linker:       %s' % linker)


# --------------------------------------------------------------------


class Extension(_Extension):
    pass


# --------------------------------------------------------------------

cmd_petsc_opts = [
    ('petsc-dir=', None, 'define PETSC_DIR, overriding environmental variables'),
    ('petsc-arch=', None, 'define PETSC_ARCH, overriding environmental variables'),
]


class config(_config):
    Configure = PetscConfig

    user_options = _config.user_options + cmd_petsc_opts

    def initialize_options(self):
        _config.initialize_options(self)
        self.petsc_dir = None
        self.petsc_arch = None

    def get_config_arch(self, arch):
        return config.Configure(self.petsc_dir, arch)

    def run(self):
        _config.run(self)
        self.petsc_dir = config.get_petsc_dir(self.petsc_dir)
        if self.petsc_dir is None:
            return
        petsc_arch = config.get_petsc_arch(self.petsc_dir, self.petsc_arch)
        log.info('-' * 70)
        log.info('PETSC_DIR:   %s' % self.petsc_dir)
        arch_list = petsc_arch
        if not arch_list:
            arch_list = [None]
        for arch in arch_list:
            conf = self.get_config_arch(arch)
            archname = conf.PETSC_ARCH or conf['PETSC_ARCH']
            scalar_type = conf['PETSC_SCALAR']
            precision = conf['PETSC_PRECISION']
            language = conf['PETSC_LANGUAGE']
            compiler = conf['PCC']
            linker = conf['PCC_LINKER']
            log.info('-' * 70)
            log.info('PETSC_ARCH:  %s' % archname)
            log.info(' * scalar-type: %s' % scalar_type)
            log.info(' * precision:   %s' % precision)
            log.info(' * language:    %s' % language)
            log.info(' * compiler:    %s' % compiler)
            log.info(' * linker:      %s' % linker)
        log.info('-' * 70)

    # @staticmethod
    def get_petsc_dir(petsc_dir):
        if not petsc_dir:
            return None
        petsc_dir = os.path.expandvars(petsc_dir)
        if not petsc_dir or '$PETSC_DIR' in petsc_dir:
            try:
                import petsc

                petsc_dir = petsc.get_petsc_dir()
            except ImportError:
                log.warn('PETSC_DIR not specified')
                return None
        petsc_dir = os.path.expanduser(petsc_dir)
        petsc_dir = os.path.abspath(petsc_dir)
        return config.chk_petsc_dir(petsc_dir)

    get_petsc_dir = staticmethod(get_petsc_dir)

    # @staticmethod
    def chk_petsc_dir(petsc_dir):
        if not os.path.isdir(petsc_dir):
            log.error('invalid PETSC_DIR: %s (ignored)' % petsc_dir)
            return None
        return petsc_dir

    chk_petsc_dir = staticmethod(chk_petsc_dir)

    # @staticmethod
    def get_petsc_arch(petsc_dir, petsc_arch):
        if not petsc_dir:
            return None
        petsc_arch = os.path.expandvars(petsc_arch)
        if not petsc_arch or '$PETSC_ARCH' in petsc_arch:
            petsc_arch = ''
            petsc_conf = os.path.join(petsc_dir, 'lib', 'petsc', 'conf')
            if os.path.isdir(petsc_conf):
                petscvariables = os.path.join(petsc_conf, 'petscvariables')
                if os.path.exists(petscvariables):
                    conf = makefile(open(petscvariables, 'rt'))
                    petsc_arch = conf.get('PETSC_ARCH', '')
        petsc_arch = petsc_arch.split(os.pathsep)
        petsc_arch = unique(petsc_arch)
        petsc_arch = [arch for arch in petsc_arch if arch]
        return config.chk_petsc_arch(petsc_dir, petsc_arch)

    get_petsc_arch = staticmethod(get_petsc_arch)

    # @staticmethod
    def chk_petsc_arch(petsc_dir, petsc_arch):
        valid_archs = []
        for arch in petsc_arch:
            arch_path = os.path.join(petsc_dir, arch)
            if os.path.isdir(arch_path):
                valid_archs.append(arch)
            else:
                log.warn('invalid PETSC_ARCH: %s (ignored)' % arch)
        return valid_archs

    chk_petsc_arch = staticmethod(chk_petsc_arch)


class build(_build):
    user_options = _build.user_options
    user_options += [
        (
            'inplace',
            'i',
            'ignore build-lib and put compiled extensions into the source '
            'directory alongside your pure Python modules',
        )
    ]
    user_options += cmd_petsc_opts

    boolean_options = _build.boolean_options
    boolean_options += ['inplace']

    def initialize_options(self):
        _build.initialize_options(self)
        self.inplace = None
        self.petsc_dir = None
        self.petsc_arch = None

    def finalize_options(self):
        _build.finalize_options(self)
        if self.inplace is None:
            self.inplace = False
        self.set_undefined_options(
            'config', ('petsc_dir', 'petsc_dir'), ('petsc_arch', 'petsc_arch')
        )
        self.petsc_dir = config.get_petsc_dir(self.petsc_dir)
        self.petsc_arch = config.get_petsc_arch(self.petsc_dir, self.petsc_arch)

    sub_commands = [('build_src', lambda *args: True)] + _build.sub_commands


class build_src(Command):
    description = 'build C sources from Cython files'

    user_options = [
        ('force', 'f', 'forcibly build everything (ignore file timestamps)'),
    ]

    boolean_options = ['force']

    def initialize_options(self):
        self.force = False

    def finalize_options(self):
        self.set_undefined_options(
            'build',
            ('force', 'force'),
        )

    def run(self):
        sources = getattr(self, 'sources', [])
        for source in sources:
            cython_run(force=self.force, VERSION=cython_req(), **source)


class build_ext(_build_ext):
    user_options = _build_ext.user_options + cmd_petsc_opts

    def initialize_options(self):
        _build_ext.initialize_options(self)
        self.inplace = None
        self.petsc_dir = None
        self.petsc_arch = None
        self._outputs = []

    def finalize_options(self):
        _build_ext.finalize_options(self)
        self.set_undefined_options('build', ('inplace', 'inplace'))
        self.set_undefined_options(
            'build', ('petsc_dir', 'petsc_dir'), ('petsc_arch', 'petsc_arch')
        )

    def _copy_ext(self, ext):
        extclass = ext.__class__
        fullname = self.get_ext_fullname(ext.name)
        modpath = str.split(fullname, '.')
        pkgpath = os.path.join('', *modpath[0:-1])
        name = modpath[-1]
        sources = list(ext.sources)
        newext = extclass(name, sources)
        newext.__dict__.update(copy.deepcopy(ext.__dict__))
        newext.name = name
        return pkgpath, newext

    def _build_ext_arch(self, ext, pkgpath, arch):
        build_temp = self.build_temp
        build_lib = self.build_lib
        try:
            self.build_temp = os.path.join(build_temp, arch)
            self.build_lib = os.path.join(build_lib, pkgpath, arch)
            _build_ext.build_extension(self, ext)
        finally:
            self.build_temp = build_temp
            self.build_lib = build_lib

    def get_config_arch(self, arch):
        return config.Configure(self.petsc_dir, arch)

    def build_extension(self, ext):
        if not isinstance(ext, Extension):
            return _build_ext.build_extension(self, ext)
        petsc_arch = self.petsc_arch
        if not petsc_arch:
            petsc_arch = [None]
        for arch in petsc_arch:
            config = self.get_config_arch(arch)
            ARCH = arch or config['PETSC_ARCH']
            if ARCH not in self.PETSC_ARCH_LIST:
                self.PETSC_ARCH_LIST.append(ARCH)
            self.DESTDIR = config.DESTDIR
            ext.language = config.language
            config.log_info()
            pkgpath, newext = self._copy_ext(ext)
            config.configure(newext, self.compiler)
            self._build_ext_arch(newext, pkgpath, ARCH)
        return None

    def run(self):
        self.build_sources()
        _build_ext.run(self)

    def build_sources(self):
        if 'build_src' in self.distribution.cmdclass:
            self.run_command('build_src')

    def build_extensions(self, *args, **kargs):
        self.PETSC_ARCH_LIST = []
        _build_ext.build_extensions(self, *args, **kargs)
        if not self.PETSC_ARCH_LIST:
            return
        self.build_configuration(self.PETSC_ARCH_LIST)

    def build_configuration(self, arch_list):
        #
        template, variables = self.get_config_data(arch_list)
        config_data = template % variables
        #
        build_lib = self.build_lib
        dist_name = self.distribution.get_name()
        config_file = os.path.join(
            build_lib, dist_name, 'lib', dist_name.replace('4py', '') + '.cfg'
        )

        #
        def write_file(filename, data):
            with open(filename, 'w') as fh:
                fh.write(config_data)

        execute(
            write_file,
            (config_file, config_data),
            msg='writing %s' % config_file,
            verbose=self.verbose,
        )

    def get_config_data(self, arch_list):
        DESTDIR = self.DESTDIR
        template = (
            '\n'.join(
                [
                    'PETSC_DIR  = %(PETSC_DIR)s',
                    'PETSC_ARCH = %(PETSC_ARCH)s',
                ]
            )
            + '\n'
        )
        variables = {
            'PETSC_DIR': strip_prefix(DESTDIR, self.petsc_dir),
            'PETSC_ARCH': os.path.pathsep.join(arch_list),
        }
        return template, variables

    def copy_extensions_to_source(self):
        build_py = self.get_finalized_command('build_py')
        for ext in self.extensions:
            inp_file, reg_file = self._get_inplace_equivalent(build_py, ext)

            arch_list = ['']
            if isinstance(ext, Extension) and self.petsc_arch:
                arch_list = self.petsc_arch[:]

            file_pairs = []
            inp_head, inp_tail = os.path.split(inp_file)
            reg_head, reg_tail = os.path.split(reg_file)
            for arch in arch_list:
                inp_file = os.path.join(inp_head, arch, inp_tail)
                reg_file = os.path.join(reg_head, arch, reg_tail)
                file_pairs.append((inp_file, reg_file))

            for inp_file, reg_file in file_pairs:
                if os.path.exists(reg_file) or not ext.optional:
                    dest_dir, _ = os.path.split(inp_file)
                    self.mkpath(dest_dir)
                    self.copy_file(reg_file, inp_file, level=self.verbose)

    def get_outputs(self):
        self.check_extensions_list(self.extensions)
        outputs = []
        for ext in self.extensions:
            fullname = self.get_ext_fullname(ext.name)
            filename = self.get_ext_filename(fullname)
            if isinstance(ext, Extension) and self.petsc_arch:
                head, tail = os.path.split(filename)
                for arch in self.petsc_arch:
                    outfile = os.path.join(self.build_lib, head, arch, tail)
                    outputs.append(outfile)
            else:
                outfile = os.path.join(self.build_lib, filename)
                outputs.append(outfile)
        return list(set(outputs))

    def get_source_files(self):
        orig = log.set_threshold(log.WARN)
        try:
            return super().get_source_files()
        finally:
            log.set_threshold(orig)


class install(_install):
    def initialize_options(self):
        with warnings.catch_warnings():
            if setuptools:
                if hasattr(setuptools, 'SetuptoolsDeprecationWarning'):
                    category = setuptools.SetuptoolsDeprecationWarning
                    warnings.simplefilter('ignore', category)
            _install.initialize_options(self)
        self.old_and_unmanageable = True


cmdclass_list = [
    config,
    build,
    build_src,
    build_ext,
    install,
]

# --------------------------------------------------------------------


def setup(**attrs):
    cmdclass = attrs.setdefault('cmdclass', {})
    for cmd in cmdclass_list:
        cmdclass.setdefault(cmd.__name__, cmd)
    build_src.sources = attrs.pop('cython_sources', None)
    use_setup_requires = False  # handle Cython requirement ourselves
    if setuptools and build_src.sources and use_setup_requires:
        version = cython_req()
        if not cython_chk(version, verbose=False):
            reqs = attrs.setdefault('setup_requires', [])
            reqs += ['Cython>=' + version]
    return _setup(**attrs)


# --------------------------------------------------------------------

if setuptools:
    try:
        from setuptools.command import egg_info as mod_egg_info

        _FileList = mod_egg_info.FileList

        class FileList(_FileList):
            def process_template_line(self, line):
                level = log.set_threshold(log.ERROR)
                try:
                    _FileList.process_template_line(self, line)
                finally:
                    log.set_threshold(level)

        mod_egg_info.FileList = FileList
    except (ImportError, AttributeError):
        pass

# --------------------------------------------------------------------


def append(seq, item):
    if item not in seq:
        seq.append(item)


def append_dict(conf, dct):
    for key, values in dct.items():
        if key in conf:
            for value in values:
                if value not in conf[key]:
                    conf[key].append(value)


def unique(seq):
    res = []
    for item in seq:
        if item not in res:
            res.append(item)
    return res


def flaglist(flags):
    conf = {
        'define_macros': [],
        'undef_macros': [],
        'include_dirs': [],
        'libraries': [],
        'library_dirs': [],
        'runtime_library_dirs': [],
        'extra_compile_args': [],
        'extra_link_args': [],
    }

    if isinstance(flags, str):
        flags = flags.split()

    switch = '-Wl,'
    newflags = []
    linkopts = []
    for f in flags:
        if f.startswith(switch):
            if len(f) > 4:
                append(linkopts, f[4:])
        else:
            append(newflags, f)
    if linkopts:
        newflags.append(switch + ','.join(linkopts))
    flags = newflags

    append_next_word = None

    for word in flags:
        if append_next_word is not None:
            append(append_next_word, word)
            append_next_word = None
            continue

        switch, value = word[0:2], word[2:]

        if switch == '-I':
            append(conf['include_dirs'], value)
        elif switch == '-D':
            try:
                idx = value.index('=')
                macro = (value[:idx], value[idx + 1 :])
            except ValueError:
                macro = (value, None)
            append(conf['define_macros'], macro)
        elif switch == '-U':
            append(conf['undef_macros'], value)
        elif switch == '-l':
            append(conf['libraries'], value)
        elif switch == '-L':
            append(conf['library_dirs'], value)
        elif switch == '-R':
            append(conf['runtime_library_dirs'], value)
        elif word.startswith('-Wl'):
            linkopts = word.split(',')
            append_dict(conf, flaglist(linkopts[1:]))
        elif word == '-rpath':
            append_next_word = conf['runtime_library_dirs']
        elif word == '-Xlinker':
            append_next_word = conf['extra_link_args']
        else:
            # log.warn("unrecognized flag '%s'" % word)
            pass
    return conf


def prepend_to_flags(path, flags):
    """Prepend a path to compiler flags with absolute paths"""
    if not path:
        return flags

    def append_path(m):
        switch = m.group(1)
        open_quote = m.group(4)
        old_path = m.group(5)
        close_quote = m.group(6)
        if os.path.isabs(old_path):
            moded_path = os.path.normpath(path + os.path.sep + old_path)
            return switch + open_quote + moded_path + close_quote
        return m.group(0)

    return re.sub(r'((^|\s+)(-I|-L))(\s*["\']?)(\S+)(["\']?)', append_path, flags)


def strip_prefix(prefix, string):
    if not prefix:
        return string
    return re.sub(r'^' + prefix, '', string)


# --------------------------------------------------------------------

# Regexes needed for parsing Makefile-like syntaxes
_variable_rx = re.compile(r'([a-zA-Z][a-zA-Z0-9_]+)\s*=\s*(.*)')
_findvar1_rx = re.compile(r'\$\(([A-Za-z][A-Za-z0-9_]*)\)')
_findvar2_rx = re.compile(r'\${([A-Za-z][A-Za-z0-9_]*)}')


def makefile(fileobj, dct=None):
    """Parse a Makefile-style file.

    A dictionary containing name/value pairs is returned.  If an
    optional dictionary is passed in as the second argument, it is
    used instead of a new dictionary.
    """
    fp = TextFile(file=fileobj, strip_comments=1, skip_blanks=1, join_lines=1)

    if dct is None:
        dct = {}
    done = {}
    notdone = {}

    while 1:
        line = fp.readline()
        if line is None:  # eof
            break
        m = _variable_rx.match(line)
        if m:
            n, v = m.group(1, 2)
            v = str.strip(v)
            if '$' in v:
                notdone[n] = v
            else:
                try:
                    v = int(v)
                except ValueError:
                    pass
                done[n] = v
                try:
                    del notdone[n]
                except KeyError:
                    pass
    fp.close()

    # do variable interpolation here
    while notdone:
        for name in list(notdone.keys()):
            value = notdone[name]
            m = _findvar1_rx.search(value) or _findvar2_rx.search(value)
            if m:
                n = m.group(1)
                found = True
                if n in done:
                    item = str(done[n])
                elif n in notdone:
                    # get it on a subsequent round
                    found = False
                else:
                    done[n] = item = ''
                if found:
                    after = value[m.end() :]
                    value = value[: m.start()] + item + after
                    if '$' in after:
                        notdone[name] = value
                    else:
                        try:
                            value = int(value)
                        except ValueError:
                            done[name] = str.strip(value)
                        else:
                            done[name] = value
                        del notdone[name]
            else:
                # bogus variable reference;
                # just drop it since we can't deal
                del notdone[name]
    # save the results in the global dictionary
    dct.update(done)
    return dct


# --------------------------------------------------------------------



# --------------------------------------------------------------------

class AOType(object):
    """The application ordering types."""
    BASIC          = S_(AOBASIC)
    ADVANCED       = S_(AOADVANCED)
    MAPPING        = S_(AOMAPPING)
    MEMORYSCALABLE = S_(AOMEMORYSCALABLE)

# --------------------------------------------------------------------


cdef class AO(Object):
    """Application ordering object."""
    Type = AOType

    def __cinit__(self):
        self.obj = <PetscObject*> &self.ao
        self.ao = NULL

    def view(self, Viewer viewer=None) -> None:
        """Display the application ordering.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` to display the ordering.

        See Also
        --------
        petsc.AOView

        """
        cdef PetscViewer cviewer = NULL
        if viewer is not None: cviewer = viewer.vwr
        CHKERR(AOView(self.ao, cviewer))

    def destroy(self) -> Self:
        """Destroy the application ordering.

        Collective.

        See Also
        --------
        petsc.AODestroy

        """
        CHKERR(AODestroy(&self.ao))
        return self

    def createBasic(
        self,
        app: Sequence[int] | IS,
        petsc: Sequence[int] | IS | None = None,
        comm: Comm | None = None) -> Self:
        """Return a basic application ordering using two orderings.

        Collective.

        The arrays/indices ``app`` and ``petsc`` must contain all the integers
        ``0`` to ``len(app)-1`` with no duplicates; that is there cannot be any
        "holes" in the indices. Use ``createMapping`` if you wish to have
        "holes" in the indices.

        Parameters
        ----------
        app
            The application ordering.
        petsc
            Another ordering (may be `None` to indicate the natural ordering,
            that is 0, 1, 2, 3, ...).
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        createMemoryScalable, createMapping, petsc.AOCreateBasicIS
        petsc.AOCreateBasic

        """
        cdef PetscIS isapp = NULL, ispetsc = NULL
        cdef PetscInt napp = 0, *idxapp = NULL,
        cdef PetscInt npetsc = 0, *idxpetsc = NULL
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscAO newao = NULL
        if isinstance(app, IS):
            isapp = (<IS>app).iset
            if petsc is not None:
                ispetsc = (<IS?>petsc).iset
            CHKERR(AOCreateBasicIS(isapp, ispetsc, &newao))
        else:
            app = iarray_i(app, &napp, &idxapp)
            if petsc is not None:
                petsc = iarray_i(petsc, &npetsc, &idxpetsc)
                assert napp == npetsc, "incompatible array sizes"
            CHKERR(AOCreateBasic(ccomm, napp, idxapp, idxpetsc, &newao))
        CHKERR(PetscCLEAR(self.obj)); self.ao = newao
        return self

    def createMemoryScalable(
        self,
        app: Sequence[int] | IS,
        petsc: Sequence[int] | IS | None = None,
        comm: Comm | None = None) -> Self:
        """Return a memory scalable application ordering using two orderings.

        Collective.

        The arrays/indices ``app`` and ``petsc`` must contain all the integers
        ``0`` to ``len(app)-1`` with no duplicates; that is there cannot be any
        "holes" in the indices. Use ``createMapping`` if you wish to have
        "holes" in the indices.

        Comparing with ``createBasic``, this routine trades memory with message
        communication.

        Parameters
        ----------
        app
            The application ordering.
        petsc
            Another ordering (may be `None` to indicate the natural ordering,
            that is 0, 1, 2, 3, ...).
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        createBasic, createMapping, petsc.AOCreateMemoryScalableIS
        petsc.AOCreateMemoryScalable

        """
        cdef PetscIS isapp = NULL, ispetsc = NULL
        cdef PetscInt napp = 0, *idxapp = NULL,
        cdef PetscInt npetsc = 0, *idxpetsc = NULL
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscAO newao = NULL
        if isinstance(app, IS):
            isapp = (<IS>app).iset
            if petsc is not None:
                ispetsc = (<IS?>petsc).iset
            CHKERR(AOCreateMemoryScalableIS(isapp, ispetsc, &newao))
        else:
            app = iarray_i(app, &napp, &idxapp)
            if petsc is not None:
                petsc = iarray_i(petsc, &npetsc, &idxpetsc)
                assert napp == npetsc, "incompatible array sizes"
            CHKERR(AOCreateMemoryScalable(ccomm, napp, idxapp, idxpetsc, &newao))
        CHKERR(PetscCLEAR(self.obj)); self.ao = newao
        return self

    def createMapping(
        self,
        app: Sequence[int] | IS,
        petsc: Sequence[int] | IS | None = None,
        comm: Comm | None = None) -> Self:
        """Return an application mapping using two orderings.

        Collective.

        The arrays ``app`` and ``petsc`` need NOT contain all the integers
        ``0`` to ``len(app)-1``, that is there CAN be "holes" in the indices.
        Use ``createBasic`` if they do not have holes for better performance.

        Parameters
        ----------
        app
            The application ordering.
        petsc
            Another ordering. May be `None` to indicate the identity ordering.
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        createBasic, petsc.AOCreateMappingIS, petsc.AOCreateMapping

        """
        cdef PetscIS isapp = NULL, ispetsc = NULL
        cdef PetscInt napp = 0, *idxapp = NULL,
        cdef PetscInt npetsc = 0, *idxpetsc = NULL
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscAO newao = NULL
        if isinstance(app, IS):
            isapp = (<IS>app).iset
            if petsc is not None:
                ispetsc = (<IS?>petsc).iset
            CHKERR(AOCreateMappingIS(isapp, ispetsc, &newao))
        else:
            app = iarray_i(app, &napp, &idxapp)
            if petsc is not None:
                petsc = iarray_i(petsc, &npetsc, &idxpetsc)
                assert napp == npetsc, "incompatible array sizes"
            CHKERR(AOCreateMapping(ccomm, napp, idxapp, idxpetsc, &newao))
        CHKERR(PetscCLEAR(self.obj)); self.ao = newao
        return self

    def getType(self) -> str:
        """Return the application ordering type.

        Not collective.

        See Also
        --------
        petsc.AOGetType

        """
        cdef PetscAOType cval = NULL
        CHKERR(AOGetType(self.ao, &cval))
        return bytes2str(cval)

    def app2petsc(self, indices: Sequence[int] | IS) -> Sequence[int] | IS:
        """Map an application-defined ordering to the PETSc ordering.

        Collective.

        Any integers in ``indices`` that are negative are left unchanged. This
        allows one to convert, for example, neighbor lists that use negative
        entries to indicate nonexistent neighbors due to boundary conditions,
        etc.

        Integers that are out of range are mapped to -1.

        If ``IS`` is used, it cannot be of type stride or block.

        Parameters
        ----------
        indices
            The indices; to be replaced with their mapped values.

        See Also
        --------
        petsc2app, petsc.AOApplicationToPetscIS, petsc.AOApplicationToPetsc

        """
        cdef PetscIS iset = NULL
        cdef PetscInt nidx = 0, *idx = NULL
        if isinstance(indices, IS):
            iset = (<IS>indices).iset
            CHKERR(AOApplicationToPetscIS(self.ao, iset))
        else:
            indices = oarray_i(indices, &nidx, &idx)
            CHKERR(AOApplicationToPetsc(self.ao, nidx, idx))
        return indices

    def petsc2app(self, indices: Sequence[int] | IS) -> Sequence[int] | IS:
        """Map a PETSc ordering to the application-defined ordering.

        Collective.

        Any integers in ``indices`` that are negative are left unchanged. This
        allows one to convert, for example, neighbor lists that use negative
        entries to indicate nonexistent neighbors due to boundary conditions,
        etc.

        Integers that are out of range are mapped to -1.

        If ``IS`` is used, it cannot be of type stride or block.

        Parameters
        ----------
        indices
            The indices; to be replaced with their mapped values.

        See Also
        --------
        app2petsc, petsc.AOPetscToApplicationIS, petsc.AOPetscToApplication

        """
        cdef PetscIS iset = NULL
        cdef PetscInt nidx = 0, *idx = NULL
        if isinstance(indices, IS):
            iset = (<IS>indices).iset
            CHKERR(AOPetscToApplicationIS(self.ao, iset))
        else:
            indices = oarray_i(indices, &nidx, &idx)
            CHKERR(AOPetscToApplication(self.ao, nidx, idx))
        return indices

# --------------------------------------------------------------------

del AOType

# --------------------------------------------------------------------



# ------------------------------------------------------------------------
#
#  Solid Fuel Ignition (SFI) problem.  This problem is modeled by the
#  partial differential equation
#
#          -Laplacian(u) - lambda * exp(u) = 0,  0 < x,y,z < 1,
#
#  with boundary conditions
#
#           u = 0  for  x = 0, x = 1, y = 0, y = 1, z = 0, z = 1
#
#  A finite difference approximation with the usual 7-point stencil
#  is used to discretize the boundary value problem to obtain a
#  nonlinear system of equations. The problem is solved in a 3D
#  rectangular domain, using distributed arrays (DAs) to partition
#  the parallel grid.
#
# ------------------------------------------------------------------------

try: range = xrange
except: pass

import sys, petsc4py
petsc4py.init(sys.argv)

from numpy import exp, sqrt
from petsc4py import PETSc

class Bratu3D(object):

    def __init__(self, da, lambda_):
        assert da.getDim() == 3
        self.da = da
        self.lambda_ = lambda_
        self.localX  = da.createLocalVec()

    def formInitGuess(self, snes, X):
        #
        x = self.da.getVecArray(X)
        #
        mx, my, mz = self.da.getSizes()
        hx, hy, hz = [1.0/(m-1) for m in [mx, my, mz]]
        lambda_ = self.lambda_
        scale = lambda_/(lambda_ + 1.0)
        #
        (xs, xe), (ys, ye), (zs, ze) = self.da.getRanges()
        for k in range(zs, ze):
            min_k = min(k,mz-k-1)*hz
            for j in range(ys, ye):
                min_j = min(j,my-j-1)*hy
                for i in range(xs, xe):
                    min_i = min(i,mx-i-1)*hx
                    if (i==0    or j==0    or k==0 or
                        i==mx-1 or j==my-1 or k==mz-1):
                        # boundary points
                        x[i, j, k] = 0.0
                    else:
                        # interior points
                        min_kij = min(min_i,min_j,min_k)
                        x[i, j, k] = scale*sqrt(min_kij)

    def formFunction(self, snes, X, F):
        #
        self.da.globalToLocal(X, self.localX)
        x = self.da.getVecArray(self.localX)
        f = self.da.getVecArray(F)
        #
        mx, my, mz = self.da.getSizes()
        hx, hy, hz = [1.0/m for m in [mx, my, mz]]
        hxhyhz  = hx*hy*hz
        hxhzdhy = hx*hz/hy;
        hyhzdhx = hy*hz/hx;
        hxhydhz = hx*hy/hz;
        lambda_ = self.lambda_
        #
        (xs, xe), (ys, ye), (zs, ze) = self.da.getRanges()
        for k in range(zs, ze):
            for j in range(ys, ye):
                for i in range(xs, xe):
                    if (i==0    or j==0    or k==0 or
                        i==mx-1 or j==my-1 or k==mz-1):
                        f[i, j, k] = x[i, j, k] - 0
                    else:
                        u   = x[ i  ,  j  ,  k ] # center
                        u_e = x[i+1 ,  j  ,  k ] # east
                        u_w = x[i-1 ,  j  ,  k ] # west
                        u_n = x[ i  , j+1 ,  k ] # north
                        u_s = x[ i  , j-1 ,  k ] # south
                        u_u = x[ i  ,  j  , k+1] # up
                        u_d = x[ i  ,  j  , k-1] # down
                        u_xx = (-u_e + 2*u - u_w)*hyhzdhx
                        u_yy = (-u_n + 2*u - u_s)*hxhzdhy
                        u_zz = (-u_u + 2*u - u_d)*hxhydhz
                        f[i, j, k] = u_xx + u_yy + u_zz \
                                     - lambda_*exp(u)*hxhyhz

    def formJacobian(self, snes, X, J, P):
        #
        self.da.globalToLocal(X, self.localX)
        x = self.da.getVecArray(self.localX)
        #
        mx, my, mz = self.da.getSizes()
        hx, hy, hz = [1.0/m for m in [mx, my, mz]]
        hxhyhz  = hx*hy*hz
        hxhzdhy = hx*hz/hy;
        hyhzdhx = hy*hz/hx;
        hxhydhz = hx*hy/hz;
        lambda_ = self.lambda_
        #
        P.zeroEntries()
        row = PETSc.Mat.Stencil()
        col = PETSc.Mat.Stencil()
        #
        (xs, xe), (ys, ye), (zs, ze) = self.da.getRanges()
        for k in range(zs, ze):
            for j in range(ys, ye):
                for i in range(xs, xe):
                    row.index = (i,j,k)
                    row.field = 0
                    if (i==0    or j==0    or k==0 or
                        i==mx-1 or j==my-1 or k==mz-1):
                        P.setValueStencil(row, row, 1.0)
                    else:
                        u = x[i,j,k]
                        diag = (2*(hyhzdhx+hxhzdhy+hxhydhz)
                                - lambda_*exp(u)*hxhyhz)
                        for index, value in [
                            ((i,j,k-1), -hxhydhz),
                            ((i,j-1,k), -hxhzdhy),
                            ((i-1,j,k), -hyhzdhx),
                            ((i, j, k), diag),
                            ((i+1,j,k), -hyhzdhx),
                            ((i,j+1,k), -hxhzdhy),
                            ((i,j,k+1), -hxhydhz),
                            ]:
                            col.index = index
                            col.field = 0
                            P.setValueStencil(row, col, value)
        P.assemble()
        if J != P: J.assemble() # matrix-free operator
        return PETSc.Mat.Structure.SAME_NONZERO_PATTERN

OptDB = PETSc.Options()

n  = OptDB.getInt('n', 16)
nx = OptDB.getInt('nx', n)
ny = OptDB.getInt('ny', n)
nz = OptDB.getInt('nz', n)
lambda_ = OptDB.getReal('lambda', 6.0)

da = PETSc.DMDA().create([nx, ny, nz], stencil_width=1)
pde = Bratu3D(da, lambda_)

snes = PETSc.SNES().create()
F = da.createGlobalVec()
snes.setFunction(pde.formFunction, F)

fd = OptDB.getBool('fd', False)
mf = OptDB.getBool('mf', False)
if mf:
    J = None
    snes.setUseMF()
else:
    J = da.createMat()
    snes.setJacobian(pde.formJacobian, J)
    if fd:
        snes.setUseFD()

snes.getKSP().setType('cg')
snes.setFromOptions()

X = da.createGlobalVec()
pde.formInitGuess(snes, X)
snes.solve(None, X)

U = da.createNaturalVec()
da.globalToNatural(X, U)

if OptDB.getBool('plot_mpl', False):

    def plot_mpl(da, U):
        comm = da.getComm()
        rank = comm.getRank()
        scatter, U0 = PETSc.Scatter.toZero(U)
        scatter.scatter(U, U0, False, PETSc.Scatter.Mode.FORWARD)
        if rank == 0:
            try:
                from matplotlib import pylab
            except ImportError:
                PETSc.Sys.Print("matplotlib not available")
            else:
                from numpy import mgrid
                nx, ny, nz = da.sizes
                solution = U0[...].reshape(da.sizes, order='f')
                xx, yy =  mgrid[0:1:1j*nx,0:1:1j*ny]
                pylab.contourf(xx, yy, solution[:, :, nz//2])
                pylab.axis('equal')
                pylab.xlabel('X')
                pylab.ylabel('Y')
                pylab.title('Z/2')
                pylab.show()
        comm.barrier()

    plot_mpl(da, U)



import unittest
from petsc4py import PETSc

# --------------------------------------------------------------------


def allclose(seq1, seq2):
    for v1, v2 in zip(seq1, seq2):
        if abs(v1 - v2) > 1e-5:
            return False
    return True


class TestNullSpace(unittest.TestCase):
    def setUp(self):
        u1 = PETSc.Vec().createSeq(3)
        u2 = PETSc.Vec().createSeq(3)
        u1[0], u1[1], u1[2] = [1, 2, 0]
        u1.normalize()
        u2[0], u2[1], u2[2] = [2, -1, 0]
        u2.normalize()
        basis = [u1, u2]
        nullsp = PETSc.NullSpace().create(False, basis, comm=PETSc.COMM_SELF)
        self.basis = basis
        self.nullsp = nullsp

    def tearDown(self):
        self.basis = None
        self.nullsp = None
        PETSc.garbage_cleanup()

    def _remove(self):
        v = PETSc.Vec().createSeq(3)
        v[0], v[1], v[2] = [7, 8, 9]
        w = v.copy()
        self.nullsp.remove(w)
        return (v, w)

    def testRemove(self):
        v, w = self._remove()
        self.assertTrue(allclose(v.array, [7, 8, 9]))
        self.assertTrue(allclose(w.array, [0, 0, 9]))
        del v, w

    def testRemoveInplace(self):
        v, w = self._remove()
        self.nullsp.remove(v)
        self.assertTrue(v.equal(w))
        del v, w

    def testRemoveWithFunction(self):
        def myremove(nsp, vec):
            vec.setArray([1, 2, 3])

        self.nullsp.setFunction(myremove)
        v, w = self._remove()
        self.assertTrue(allclose(v.array, [7, 8, 9]))
        self.assertTrue(allclose(w.array, [1, 2, 3]))
        self.nullsp.remove(v)
        self.assertTrue(allclose(v.array, [1, 2, 3]))
        self.nullsp.setFunction(None)
        self.testRemove()

    def testGetSetFunction(self):
        def rem(nsp, vec):
            vec.set(0)

        self.nullsp.setFunction(rem)
        dct = self.nullsp.getDict()
        self.assertTrue(dct is not None)
        fun, a, kw = dct['__function__']
        self.assertTrue(fun is rem)
        self.nullsp.setFunction(None)
        fun = dct.get('__function__')
        self.assertTrue(fun is None)


# --------------------------------------------------------------------

if __name__ == '__main__':
    unittest.main()

# --------------------------------------------------------------------



# --------------------------------------------------------------------

class DMSwarmType(object):
    """Swarm types."""
    BASIC = DMSWARM_BASIC
    PIC = DMSWARM_PIC


class DMSwarmMigrateType(object):
    """Swarm migration types."""
    MIGRATE_BASIC = DMSWARM_MIGRATE_BASIC
    MIGRATE_DMCELLNSCATTER = DMSWARM_MIGRATE_DMCELLNSCATTER
    MIGRATE_DMCELLEXACT = DMSWARM_MIGRATE_DMCELLEXACT
    MIGRATE_USER = DMSWARM_MIGRATE_USER


class DMSwarmCollectType(object):
    """Swarm collection types."""
    COLLECT_BASIC = DMSWARM_COLLECT_BASIC
    COLLECT_DMDABOUNDINGBOX = DMSWARM_COLLECT_DMDABOUNDINGBOX
    COLLECT_GENERAL = DMSWARM_COLLECT_GENERAL
    COLLECT_USER = DMSWARM_COLLECT_USER


class DMSwarmPICLayoutType(object):
    """Swarm PIC layout types."""
    LAYOUT_REGULAR = DMSWARMPIC_LAYOUT_REGULAR
    LAYOUT_GAUSS = DMSWARMPIC_LAYOUT_GAUSS
    LAYOUT_SUBDIVISION = DMSWARMPIC_LAYOUT_SUBDIVISION


cdef class DMSwarm(DM):
    """A `DM` object used to represent arrays of data (fields) of arbitrary type."""
    Type = DMSwarmType
    MigrateType = DMSwarmMigrateType
    CollectType = DMSwarmCollectType
    PICLayoutType = DMSwarmPICLayoutType

    def create(self, comm: Comm | None = None) -> Self:
        """Create an empty DM object and set its type to `DM.Type.SWARM`.

        Collective.

        DMs are the abstract objects in PETSc that mediate between meshes and
        discretizations and the algebraic solvers, time integrators, and
        optimization algorithms.

        Parameters
        ----------
        comm
            MPI communicator, defaults to `Sys.getDefaultComm`.

        See Also
        --------
        petsc.DMCreate, petsc.DMSetType

        """
        cdef MPI_Comm ccomm = def_Comm(comm, PETSC_COMM_DEFAULT)
        cdef PetscDM newdm = NULL
        CHKERR(DMCreate(ccomm, &newdm))
        CHKERR(PetscCLEAR(self.obj)); self.dm = newdm
        CHKERR(DMSetType(self.dm, DMSWARM))
        return self

    def createGlobalVectorFromField(self, fieldname: str) -> Vec:
        """Create a global `Vec` object associated with a given field.

        Collective.

        The vector must be returned to the `DMSwarm` using a matching call to
        `destroyGlobalVectorFromField`.

        Parameters
        ----------
        fieldname
            The textual name given to a registered field.

        See Also
        --------
        destroyGlobalVectorFromField, petsc.DMSwarmCreateGlobalVectorFromField

        """
        cdef const char *cfieldname = NULL
        cdef Vec vg = Vec()
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmCreateGlobalVectorFromField(self.dm, cfieldname, &vg.vec))
        return vg

    def destroyGlobalVectorFromField(self, fieldname: str) -> None:
        """Destroy the global `Vec` object associated with a given field.

        Collective.

        Parameters
        ----------
        fieldname
            The textual name given to a registered field.

        See Also
        --------
        createGlobalVectorFromField, petsc.DMSwarmDestroyGlobalVectorFromField

        """
        cdef const char *cfieldname = NULL
        cdef PetscVec vec = NULL
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmDestroyGlobalVectorFromField(self.dm, cfieldname, &vec))

    def createGlobalVectorFromFields(self, fieldnames: Sequence[str]) -> Vec:
        """Create a global `Vec` object associated with a given set of fields.

        Collective.

        The vector must be returned to the `DMSwarm` using a matching call to
        `destroyGlobalVectorFromFields`.

        Parameters
        ----------
        fieldnames
            The textual name given to each registered field.

        See Also
        --------
        destroyGlobalVectorFromFields, petsc.DMSwarmCreateGlobalVectorFromFields

        """
        cdef const char *cval = NULL
        cdef PetscInt i = 0, nf = <PetscInt> len(fieldnames)
        cdef const char **cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(nf), NULL, <void**>&cfieldnames)
        fieldnames = list(fieldnames)
        for i from 0 <= i < nf:
            str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
        cdef Vec vg = Vec()
        CHKERR(DMSwarmCreateGlobalVectorFromFields(self.dm, nf, cfieldnames, &vg.vec))
        return vg

    def destroyGlobalVectorFromFields(self, fieldnames: Sequence[str]) -> None:
        """Destroy the global `Vec` object associated with a given set of fields.

        Collective.

        Parameters
        ----------
        fieldnames
            The textual name given to each registered field.

        See Also
        --------
        createGlobalVectorFromFields, petsc.DMSwarmDestroyGlobalVectorFromFields

        """
        cdef const char *cval = NULL
        cdef PetscInt i = 0, nf = <PetscInt> len(fieldnames)
        cdef const char **cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(nf), NULL, <void**>&cfieldnames)
        fieldnames = list(fieldnames)
        for i from 0 <= i < nf:
            str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
        cdef PetscVec vec = NULL
        CHKERR(DMSwarmDestroyGlobalVectorFromFields(self.dm, nf, cfieldnames, &vec))

    def createLocalVectorFromField(self, fieldname: str) -> Vec:
        """Create a local `Vec` object associated with a given field.

        Collective.

        The vector must be returned to the `DMSwarm` using a matching call
        to `destroyLocalVectorFromField`.

        Parameters
        ----------
        fieldname
            The textual name given to a registered field.

        See Also
        --------
        destroyLocalVectorFromField, petsc.DMSwarmCreateLocalVectorFromField

        """
        cdef const char *cfieldname = NULL
        cdef Vec vl = Vec()
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmCreateLocalVectorFromField(self.dm, cfieldname, &vl.vec))
        return vl

    def destroyLocalVectorFromField(self, fieldname: str) -> None:
        """Destroy the local `Vec` object associated with a given field.

        Collective.

        Parameters
        ----------
        fieldname
            The textual name given to a registered field.

        See Also
        --------
        createLocalVectorFromField, petsc.DMSwarmDestroyLocalVectorFromField

        """
        cdef const char *cfieldname = NULL
        cdef PetscVec vec = NULL
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmDestroyLocalVectorFromField(self.dm, cfieldname, &vec))

    def createLocalVectorFromFields(self, fieldnames: Sequence[str]) -> Vec:
        """Create a local `Vec` object associated with a given set of fields.

        Collective.

        The vector must be returned to the `DMSwarm` using a matching call to
        `destroyLocalVectorFromFields`.

        Parameters
        ----------
        fieldnames
            The textual name given to each registered field.

        See Also
        --------
        destroyLocalVectorFromFields, petsc.DMSwarmCreateLocalVectorFromFields

        """
        cdef const char *cval = NULL
        cdef PetscInt i = 0, nf = <PetscInt> len(fieldnames)
        cdef const char **cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(nf), NULL, <void**>&cfieldnames)
        fieldnames = list(fieldnames)
        for i from 0 <= i < nf:
            str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
        cdef Vec vl = Vec()
        CHKERR(DMSwarmCreateLocalVectorFromFields(self.dm, nf, cfieldnames, &vl.vec))
        return vl

    def destroyLocalVectorFromFields(self, fieldnames: Sequence[str]) -> None:
        """Destroy the local `Vec` object associated with a given set of fields.

        Collective.

        Parameters
        ----------
        fieldnames
            The textual name given to each registered field.

        See Also
        --------
        createLocalVectorFromFields, petsc.DMSwarmDestroyLocalVectorFromFields

        """
        cdef const char *cval = NULL
        cdef PetscInt i = 0, nf = <PetscInt> len(fieldnames)
        cdef const char **cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(nf), NULL, <void**>&cfieldnames)
        fieldnames = list(fieldnames)
        for i from 0 <= i < nf:
            str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
        cdef PetscVec vec = NULL
        CHKERR(DMSwarmDestroyLocalVectorFromFields(self.dm, nf, cfieldnames, &vec))

    def initializeFieldRegister(self) -> None:
        """Initiate the registration of fields to a `DMSwarm`.

        Collective.

        After all fields have been registered, you must call `finalizeFieldRegister`.

        See Also
        --------
        finalizeFieldRegister, petsc.DMSwarmInitializeFieldRegister

        """
        CHKERR(DMSwarmInitializeFieldRegister(self.dm))

    def finalizeFieldRegister(self) -> None:
        """Finalize the registration of fields to a `DMSwarm`.

        Collective.

        See Also
        --------
        initializeFieldRegister, petsc.DMSwarmFinalizeFieldRegister

        """
        CHKERR(DMSwarmFinalizeFieldRegister(self.dm))

    def setLocalSizes(self, nlocal: int, buffer: int) -> Self:
        """Set the length of all registered fields on the `DMSwarm`.

        Not collective.

        Parameters
        ----------
        nlocal
            The length of each registered field.
        buffer
            The length of the buffer used for efficient dynamic resizing.

        See Also
        --------
        petsc.DMSwarmSetLocalSizes

        """
        cdef PetscInt cnlocal = asInt(nlocal)
        cdef PetscInt cbuffer = asInt(buffer)
        CHKERR(DMSwarmSetLocalSizes(self.dm, cnlocal, cbuffer))
        return self

    def registerField(self, fieldname: str, blocksize: int, dtype: type | dtype = ScalarType) -> None:
        """Register a field to a `DMSwarm` with a native PETSc data type.

        Collective.

        Parameters
        ----------
        fieldname
            The textual name to identify this field.
        blocksize
            The number of each data type.
        dtype
            A valid PETSc data type.

        See Also
        --------
        petsc.DMSwarmRegisterPetscDatatypeField

        """
        cdef const char *cfieldname = NULL
        cdef PetscInt cblocksize = asInt(blocksize)
        cdef PetscDataType ctype  = PETSC_DATATYPE_UNKNOWN
        if dtype == IntType:     ctype = PETSC_INT
        if dtype == RealType:    ctype = PETSC_REAL
        if dtype == ScalarType:  ctype = PETSC_SCALAR
        if dtype == ComplexType: ctype = PETSC_COMPLEX
        assert ctype != PETSC_DATATYPE_UNKNOWN
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmRegisterPetscDatatypeField(self.dm, cfieldname, cblocksize, ctype))

    def getField(self, fieldname: str) -> Sequence[int | float | complex]:
        """Return arrays storing all entries associated with a field.

        Not collective.

        The returned array contains underlying values of the field.

        The array must be returned to the `DMSwarm` using a matching call to
        `restoreField`.

        Parameters
        ----------
        fieldname
            The textual name to identify this field.

        Returns
        -------
        `numpy.ndarray`
            The type of the entries in the array will match the type of the
            field. The array is two dimensional with shape ``(num_points, blocksize)``.

        See Also
        --------
        restoreField, petsc.DMSwarmGetField

        """
        cdef const char *cfieldname = NULL
        cdef PetscInt blocksize = 0
        cdef PetscDataType ctype = PETSC_DATATYPE_UNKNOWN
        cdef PetscReal *data = NULL
        cdef PetscInt nlocal = 0
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmGetField(self.dm, cfieldname, &blocksize, &ctype, <void**> &data))
        CHKERR(DMSwarmGetLocalSize(self.dm, &nlocal))
        cdef int typenum = -1
        if ctype == PETSC_INT:     typenum = NPY_PETSC_INT
        if ctype == PETSC_REAL:    typenum = NPY_PETSC_REAL
        if ctype == PETSC_SCALAR:  typenum = NPY_PETSC_SCALAR
        if ctype == PETSC_COMPLEX: typenum = NPY_PETSC_COMPLEX
        assert typenum != -1
        cdef npy_intp s[2]
        s[0] = <npy_intp> nlocal
        s[1] = <npy_intp> blocksize
        return <object> PyArray_SimpleNewFromData(2, s, typenum, data)

    def restoreField(self, fieldname: str) -> None:
        """Restore accesses associated with a registered field.

        Not collective.

        Parameters
        ----------
        fieldname
            The textual name to identify this field.

        See Also
        --------
        getField, petsc.DMSwarmRestoreField

        """
        cdef const char *cfieldname = NULL
        cdef PetscInt blocksize = 0
        cdef PetscDataType ctype = PETSC_DATATYPE_UNKNOWN
        fieldname = str2bytes(fieldname, &cfieldname)
        CHKERR(DMSwarmRestoreField(self.dm, cfieldname, &blocksize, &ctype, <void**> 0))

    def vectorDefineField(self, fieldname: str) -> None:
        """Set the fields from which to define a `Vec` object.

        Collective.

        The field will be used when `DM.createLocalVec`, or
        `DM.createGlobalVec` is called.

        Parameters
        ----------
        fieldname
            The textual names given to a registered field.

        See Also
        --------
        petsc.DMSwarmVectorDefineField

        """
        cdef const char *cval = NULL
        fieldname = str2bytes(fieldname, &cval)
        CHKERR(DMSwarmVectorDefineField(self.dm, cval))

    def addPoint(self) -> None:
        """Add space for one new point in the `DMSwarm`.

        Not collective.

        See Also
        --------
        petsc.DMSwarmAddPoint

        """
        CHKERR(DMSwarmAddPoint(self.dm))

    def addNPoints(self, npoints: int) -> None:
        """Add space for a number of new points in the `DMSwarm`.

        Not collective.

        Parameters
        ----------
        npoints
            The number of new points to add.

        See Also
        --------
        petsc.DMSwarmAddNPoints

        """
        cdef PetscInt cnpoints = asInt(npoints)
        CHKERR(DMSwarmAddNPoints(self.dm, cnpoints))

    def removePoint(self) -> None:
        """Remove the last point from the `DMSwarm`.

        Not collective.

        See Also
        --------
        petsc.DMSwarmRemovePoint

        """
        CHKERR(DMSwarmRemovePoint(self.dm))

    def removePointAtIndex(self, index: int) -> None:
        """Remove a specific point from the `DMSwarm`.

        Not collective.

        Parameters
        ----------
        index
            Index of point to remove

        See Also
        --------
        petsc.DMSwarmRemovePointAtIndex

        """
        cdef PetscInt cindex = asInt(index)
        CHKERR(DMSwarmRemovePointAtIndex(self.dm, cindex))

    def copyPoint(self, pi: int, pj: int) -> None:
        """Copy point pi to point pj in the `DMSwarm`.

        Not collective.

        Parameters
        ----------
        pi
            The index of the point to copy (source).
        pj
            The point index where the copy should be located (destination).

        See Also
        --------
        petsc.DMSwarmCopyPoint

        """
        cdef PetscInt cpi = asInt(pi)
        cdef PetscInt cpj = asInt(pj)
        CHKERR(DMSwarmCopyPoint(self.dm, cpi, cpj))

    def getLocalSize(self) -> int:
        """Return the local length of fields registered.

        Not collective.

        See Also
        --------
        petsc.DMSwarmGetLocalSize

        """
        cdef PetscInt size = asInt(0)
        CHKERR(DMSwarmGetLocalSize(self.dm, &size))
        return toInt(size)

    def getSize(self) -> int:
        """Return the total length of fields registered.

        Collective.

        See Also
        --------
        petsc.DMSwarmGetSize

        """
        cdef PetscInt size = asInt(0)
        CHKERR(DMSwarmGetSize(self.dm, &size))
        return toInt(size)

    def migrate(self, remove_sent_points: bool = False) -> None:
        """Relocate points defined in the `DMSwarm` to other MPI ranks.

        Collective.

        Parameters
        ----------
        remove_sent_points
            Flag indicating if sent points should be removed from the current
            MPI rank.

        See Also
        --------
        petsc.DMSwarmMigrate

        """
        cdef PetscBool remove_pts = asBool(remove_sent_points)
        CHKERR(DMSwarmMigrate(self.dm, remove_pts))

    def collectViewCreate(self) -> None:
        """Apply a collection method and gather points in neighbor ranks.

        Collective.

        See Also
        --------
        collectViewDestroy, petsc.DMSwarmCollectViewCreate

        """
        CHKERR(DMSwarmCollectViewCreate(self.dm))

    def collectViewDestroy(self) -> None:
        """Reset the `DMSwarm` to the size prior to calling `collectViewCreate`.

        Collective.

        See Also
        --------
        collectViewCreate, petsc.DMSwarmCollectViewDestroy

        """
        CHKERR(DMSwarmCollectViewDestroy(self.dm))

    def setCellDM(self, DM dm) -> None:
        """Attach a `DM` to a `DMSwarm`.

        Collective.

        Parameters
        ----------
        dm
            The `DM` to attach to the `DMSwarm`.

        See Also
        --------
        getCellDM, petsc.DMSwarmSetCellDM

        """
        CHKERR(DMSwarmSetCellDM(self.dm, dm.dm))

    def getCellDM(self) -> DM:
        """Return `DM` cell attached to `DMSwarm`.

        Collective.

        See Also
        --------
        setCellDM, petsc.DMSwarmGetCellDM

        """
        cdef PetscDM newdm = NULL
        CHKERR(DMSwarmGetCellDM(self.dm, &newdm))
        cdef DM dm = subtype_DM(newdm)()
        dm.dm = newdm
        CHKERR(PetscINCREF(dm.obj))
        return dm

    def setType(self, dmswarm_type: Type | str) -> None:
        """Set particular flavor of `DMSwarm`.

        Collective.

        Parameters
        ----------
        dmswarm_type
            The `DMSwarm` type.

        See Also
        --------
        petsc.DMSwarmSetType

        """
        cdef PetscDMSwarmType cval = dmswarm_type
        CHKERR(DMSwarmSetType(self.dm, cval))

    def setPointsUniformCoordinates(
        self,
        min: Sequence[float],
        max: Sequence[float],
        npoints: Sequence[int],
        mode: InsertMode | None = None) -> Self:
        """Set point coordinates in a `DMSwarm` on a regular (ijk) grid.

        Collective.

        Parameters
        ----------
        min
            Minimum coordinate values in the x, y, z directions (array of
            length ``dim``).
        max
            Maximum coordinate values in the x, y, z directions (array of
            length ``dim``).
        npoints
            Number of points in each spatial direction (array of length ``dim``).
        mode
            Indicates whether to append points to the swarm (`InsertMode.ADD`),
            or override existing points (`InsertMode.INSERT`).

        See Also
        --------
        petsc.DMSwarmSetPointsUniformCoordinates

        """
        cdef PetscInt dim = asInt(0)
        CHKERR(DMGetDimension(self.dm, &dim))
        cdef PetscReal cmin[3]
        cmin[0] = cmin[1] = cmin[2] = asReal(0.)
        for i from 0 <= i < dim: cmin[i] = min[i]
        cdef PetscReal cmax[3]
        cmax[0] = cmax[1] = cmax[2] = asReal(0.)
        for i from 0 <= i < dim: cmax[i] = max[i]
        cdef PetscInt cnpoints[3]
        cnpoints[0] = cnpoints[1] = cnpoints[2] = asInt(0)
        for i from 0 <= i < dim: cnpoints[i] = npoints[i]
        cdef PetscInsertMode cmode = insertmode(mode)
        CHKERR(DMSwarmSetPointsUniformCoordinates(self.dm, cmin, cmax, cnpoints, cmode))
        return self

    def setPointCoordinates(
        self,
        coordinates: Sequence[float],
        redundant: bool = False,
        mode: InsertMode | None = None) -> None:
        """Set point coordinates in a `DMSwarm` from a user-defined list.

        Collective.

        Parameters
        ----------
        coordinates
            The coordinate values.
        redundant
            If set to `True`, it is assumed that coordinates are only valid on
            rank 0 and should be broadcast to other ranks.
        mode
            Indicates whether to append points to the swarm (`InsertMode.ADD`),
            or override existing points (`InsertMode.INSERT`).

        See Also
        --------
        petsc.DMSwarmSetPointCoordinates

        """
        cdef ndarray xyz = iarray(coordinates, NPY_PETSC_REAL)
        if PyArray_ISFORTRAN(xyz): xyz = PyArray_Copy(xyz)
        if PyArray_NDIM(xyz) != 2: raise ValueError(
            ("coordinates must have two dimensions: "
             "coordinates.ndim=%d") % (PyArray_NDIM(xyz)))
        cdef PetscInt cnpoints = <PetscInt> PyArray_DIM(xyz, 0)
        cdef PetscBool credundant = asBool(redundant)
        cdef PetscInsertMode cmode = insertmode(mode)
        cdef PetscReal *coords = <PetscReal*> PyArray_DATA(xyz)
        CHKERR(DMSwarmSetPointCoordinates(self.dm, cnpoints, coords, credundant, cmode))

    def insertPointUsingCellDM(self, layoutType: PICLayoutType, fill_param: int) -> None:
        """Insert point coordinates within each cell.

        Not collective.

        Parameters
        ----------
        layout_type
            Method used to fill each cell with the cell DM.
        fill_param
            Parameter controlling how many points per cell are added (the
            meaning of this parameter is dependent on the layout type).

        See Also
        --------
        petsc.DMSwarmInsertPointsUsingCellDM

        """
        cdef PetscDMSwarmPICLayoutType clayoutType = layoutType
        cdef PetscInt cfill_param = asInt(fill_param)
        CHKERR(DMSwarmInsertPointsUsingCellDM(self.dm, clayoutType, cfill_param))

    def setPointCoordinatesCellwise(self, coordinates: Sequence[float]) -> None:
        """Insert point coordinates within each cell.

        Not collective.

        Point coordinates are defined over the reference cell.

        Parameters
        ----------
        coordinates
            The coordinates (defined in the local coordinate system for each
            cell) to insert.

        See Also
        --------
        petsc.DMSwarmSetPointCoordinatesCellwise

        """
        cdef ndarray xyz = iarray(coordinates, NPY_PETSC_REAL)
        if PyArray_ISFORTRAN(xyz): xyz = PyArray_Copy(xyz)
        if PyArray_NDIM(xyz) != 2: raise ValueError(
            ("coordinates must have two dimensions: "
             "coordinates.ndim=%d") % (PyArray_NDIM(xyz)))
        cdef PetscInt cnpoints = <PetscInt> PyArray_DIM(xyz, 0)
        cdef PetscReal *coords = <PetscReal*> PyArray_DATA(xyz)
        CHKERR(DMSwarmSetPointCoordinatesCellwise(self.dm, cnpoints, coords))

    def viewFieldsXDMF(self, filename: str, fieldnames: Sequence[str]) -> None:
        """Write a selection of `DMSwarm` fields to an XDMF3 file.

        Collective.

        Parameters
        ----------
        filename
            The file name of the XDMF file (must have the extension .xmf).
        fieldnames
            Array containing the textual names of fields to write.

        See Also
        --------
        petsc.DMSwarmViewFieldsXDMF

        """
        cdef const char *cval = NULL
        cdef const char *cfilename = NULL
        filename = str2bytes(filename, &cfilename)
        cdef PetscInt cnfields = <PetscInt> len(fieldnames)
        cdef const char** cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(cnfields), NULL, <void**>&cfieldnames)
        fieldnames = list(fieldnames)
        for i from 0 <= i < cnfields:
            fieldnames[i] = str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
        CHKERR(DMSwarmViewFieldsXDMF(self.dm, cfilename, cnfields, cfieldnames))

    def viewXDMF(self, filename: str) -> None:
        """Write this `DMSwarm` fields to an XDMF3 file.

        Collective.

        Parameters
        ----------
        filename
            The file name of the XDMF file (must have the extension .xmf).

        See Also
        --------
        petsc.DMSwarmViewXDMF

        """
        cdef const char *cval = NULL
        filename = str2bytes(filename, &cval)
        CHKERR(DMSwarmViewXDMF(self.dm, cval))

    def sortGetAccess(self) -> None:
        """Setup up a `DMSwarm` point sort context.

        Not collective.

        The point sort context is used for efficient traversal of points within
        a cell.

        You must call `sortRestoreAccess` when you no longer need access to the
        sort context.

        See Also
        --------
        sortRestoreAccess, petsc.DMSwarmSortGetAccess

        """
        CHKERR(DMSwarmSortGetAccess(self.dm))

    def sortRestoreAccess(self) -> None:
        """Invalidate the `DMSwarm` point sorting context.

        Not collective.

        See Also
        --------
        sortGetAccess, petsc.DMSwarmSortRestoreAccess

        """
        CHKERR(DMSwarmSortRestoreAccess(self.dm))

    def sortGetPointsPerCell(self, e: int) -> list[int]:
        """Create an array of point indices for all points in a cell.

        Not collective.

        Parameters
        ----------
        e
            The index of the cell.

        See Also
        --------
        petsc.DMSwarmSortGetPointsPerCell

        """
        cdef PetscInt ce = asInt(e)
        cdef PetscInt cnpoints = asInt(0)
        cdef PetscInt *cpidlist = NULL
        cdef list pidlist = []
        CHKERR(DMSwarmSortGetPointsPerCell(self.dm, ce, &cnpoints, &cpidlist))
        npoints = asInt(cnpoints)
        for i from 0 <= i < npoints: pidlist.append(asInt(cpidlist[i]))
        return pidlist

    def sortGetNumberOfPointsPerCell(self, e: int) -> int:
        """Return the number of points in a cell.

        Not collective.

        Parameters
        ----------
        e
            The index of the cell.

        See Also
        --------
        petsc.DMSwarmSortGetNumberOfPointsPerCell

        """
        cdef PetscInt ce = asInt(e)
        cdef PetscInt npoints = asInt(0)
        CHKERR(DMSwarmSortGetNumberOfPointsPerCell(self.dm, ce, &npoints))
        return toInt(npoints)

    def sortGetIsValid(self) -> bool:
        """Return whether the sort context is up-to-date.

        Not collective.

        Returns the flag associated with a `DMSwarm` point sorting context.

        See Also
        --------
        petsc.DMSwarmSortGetIsValid

        """
        cdef PetscBool isValid = asBool(False)
        CHKERR(DMSwarmSortGetIsValid(self.dm, &isValid))
        return toBool(isValid)

    def sortGetSizes(self) -> tuple[int, int]:
        """Return the sizes associated with a `DMSwarm` point sorting context.

        Not collective.

        Returns
        -------
        ncells : int
            Number of cells within the sort context.
        npoints : int
            Number of points used to create the sort context.

        See Also
        --------
        petsc.DMSwarmSortGetSizes

        """
        cdef PetscInt ncells = asInt(0)
        cdef PetscInt npoints = asInt(0)
        CHKERR(DMSwarmSortGetSizes(self.dm, &ncells, &npoints))
        return (toInt(ncells), toInt(npoints))

    def projectFields(self, DM dm, fieldnames: Sequence[str], vecs: Sequence[Vec], mode: ScatterModeSpec = None) -> None:
        """Project a set of `DMSwarm` fields onto the cell `DM`.

        Collective.

        Parameters
        ----------
        dm
            The continuum `DM`.
        fieldnames
            The textual names of the swarm fields to project.

        See Also
        --------
        petsc.DMSwarmProjectFields

        """
        cdef const char *cval = NULL
        cdef PetscInt cnfields = <PetscInt> len(fieldnames)
        cdef const char** cfieldnames = NULL
        cdef object unused = oarray_p(empty_p(cnfields), NULL, <void**>&cfieldnames)
        cdef PetscVec *cfieldvecs = NULL
        cdef object unused2 = oarray_p(empty_p(cnfields), NULL, <void**>&cfieldvecs)
        cdef PetscScatterMode cmode = scattermode(mode)
        fieldnames = list(fieldnames)
        for i from 0 <= i < cnfields:
            fieldnames[i] = str2bytes(fieldnames[i], &cval)
            cfieldnames[i] = cval
            cfieldvecs[i] = (<Vec?>(vecs[i])).vec
        CHKERR(DMSwarmProjectFields(self.dm, dm.dm, cnfields, cfieldnames, cfieldvecs, cmode))
        return

    def addCellDM(self, CellDM celldm) -> None:
        """Add a cell DM to the `DMSwarm`.

        Logically collective.

        Parameters
        -------
        cellDM : CellDM
            The cell DM object

        See Also
        --------
        petsc.DMSwarmAddCellDM

        """
        CHKERR(DMSwarmAddCellDM(self.dm, celldm.cdm))

    def setCellDMActive(self, name : str) -> None:
        """Activate a cell DM in the `DMSwarm`.

        Logically collective.

        Parameters
        -------
        name : str
            The name of the cell DM to activate

        See Also
        --------
        petsc.DMSwarmSetCellDMActive

        """
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        CHKERR(DMSwarmSetCellDMActive(self.dm, cname))

    def getCellDMActive(self) -> CellDM:
        """Return the active cell DM in the `DMSwarm`.

        Not collective.

        See Also
        --------
        petsc.DMSwarmGetCellDMActive

        """
        cdef PetscDMSwarmCellDM newcdm = NULL
        CHKERR(DMSwarmGetCellDMActive(self.dm, &newcdm))
        cdef CellDM cdm = CellDM()
        cdm.cdm = newcdm
        CHKERR(PetscINCREF(cdm.obj))
        return cdm

    def getCellDMByName(self, name: str) -> CellDM:
        """Return the cell DM with the given name in the `DMSwarm`.

        Not collective.

        See Also
        --------
        petsc.DMSwarmGetCellDMByName

        """
        cdef PetscDMSwarmCellDM newcdm = NULL
        cdef const char *cname = NULL
        str2bytes(name, &cname)
        CHKERR(DMSwarmGetCellDMByName(self.dm, cname, &newcdm))
        cdef CellDM cdm = CellDM()
        cdm.cdm = newcdm
        CHKERR(PetscINCREF(cdm.obj))
        return cdm

    def getCellDMNames(self) -> list[str]:
        """Return the names of all cell DMs in the `DMSwarm`.

        Not collective.

        See Also
        --------
        petsc.DMSwarmGetCellDMNames

        """
        cdef PetscInt i = 0, nc = 0
        cdef const char **c = NULL
        CHKERR(DMSwarmGetCellDMNames(self.dm, &nc, &c))
        cdef list names = []
        for i from 0 <= i < nc:
            names.append(bytes2str(c[i]))
        return names

    def computeMoments(self, coord: str, weight: str) -> list[float]:
        """Return the moments defined in the active cell DM.

        Collective.

        Notes
        -----
        We integrate the given weight field over the given coordinate

        See Also
        --------
        petsc.DMSwarmComputeMoments

        """
        cdef PetscReal *mom = NULL
        cdef PetscInt cbs = 0
        cdef const char *ccoord = NULL
        cdef const char *cweight = NULL
        str2bytes(coord, &ccoord)
        str2bytes(weight, &cweight)
        CHKERR(DMSwarmGetFieldInfo(self.dm, ccoord, &cbs, NULL))
        cdef object moments = oarray_r(empty_r(asInt(cbs) + 2), NULL, &mom)
        CHKERR(DMSwarmComputeMoments(self.dm, ccoord, cweight, mom))
        return moments

# --------------------------------------------------------------------

cdef class CellDM(Object):
    """CellDM object.

    See Also
    --------
    petsc.DMSwarmCellDM

    """
    #

    def __cinit__(self):
        self.obj = <PetscObject*> &self.cdm
        self.cdm = NULL

    #

    def view(self, Viewer viewer=None) -> None:
        """View the cell DM.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        Viewer, petsc.DMSwarmCellDMView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        CHKERR(DMSwarmCellDMView(self.cdm, vwr))

    def destroy(self) -> Self:
        """Destroy the cell DM.

        Collective.

        See Also
        --------
        create, petsc.DMSwarmCellDMDestroy

        """
        CHKERR(DMSwarmCellDMDestroy(&self.cdm))
        return self

    def create(
        self,
        DM dm,
        fields: Sequence[str],
        coords: Sequence[str]) -> Self:
        """Create the cell DM.

        Collective.

        Parameters
        ----------
        dm
            The underlying DM on which to place particles
        fields
            The swarm fields represented on the DM
        coords
            The swarm fields to use as coordinates in the DM

        See Also
        --------
        destroy, petsc.DMSwarmCellDMCreate

        """
        cdef const char *cval = NULL
        cdef PetscInt i = 0, nf = <PetscInt> len(fields)
        cdef const char** fieldnames = NULL
        cdef object unuseda = oarray_p(empty_p(nf), NULL, <void**>&fieldnames)
        fields = list(fields)
        for i from 0 <= i < nf:
            str2bytes(fields[i], &cval)
            fieldnames[i] = cval
        cdef PetscInt nc = <PetscInt> len(coords)
        cdef const char** coordnames = NULL
        cdef object unusedb = oarray_p(empty_p(nf), NULL, <void**>&coordnames)
        coords = list(coords)
        for i from 0 <= i < nc:
            str2bytes(coords[i], &cval)
            coordnames[i] = cval
        cdef PetscDMSwarmCellDM newcdm = NULL
        CHKERR(DMSwarmCellDMCreate(dm.dm, nf, fieldnames, nc, coordnames, &newcdm))
        CHKERR(PetscCLEAR(self.obj)); self.cdm = newcdm
        return self

    def getDM(self) -> DM:
        """Return the underlying DM.

        Not collective.

        See Also
        --------
        petsc.DMSwarmCellDMGetDM

        """
        cdef PetscDM newdm = NULL
        CHKERR(DMSwarmCellDMGetDM(self.cdm, &newdm))
        cdef DM dm = subtype_DM(newdm)()
        dm.dm = newdm
        CHKERR(PetscINCREF(dm.obj))
        return dm

    def getCellID(self) -> str:
        """Return the cellid field for this cell DM.

        Not collective.

        See Also
        --------
        petsc.DMSwarmCellDMGetCellID

        """
        cdef const char *cellid = NULL
        CHKERR(DMSwarmCellDMGetCellID(self.cdm, &cellid))
        return bytes2str(cellid)

    def getBlockSize(self, DM sw) -> int:
        """Return the block size for this cell DM.

        Not collective.

        Parameters
        ----------
        sw:
            The `DMSwarm` object

        See Also
        --------
        petsc.DMSwarmCellDMGetBlockSize

        """
        cdef PetscInt bs = 0
        CHKERR(DMSwarmCellDMGetBlockSize(self.cdm, sw.dm, &bs))
        return toInt(bs)

    def getFields(self) -> list[str]:
        """Return the swarm fields defined on this cell DM.

        Not collective.

        See Also
        --------
        petsc.DMSwarmCellDMGetFields

        """
        cdef PetscInt i = 0, nf = 0
        cdef const char **f = NULL
        CHKERR(DMSwarmCellDMGetFields(self.cdm, &nf, &f))
        cdef list fields = []
        for i from 0 <= i < nf:
            fields.append(bytes2str(f[i]))
        return fields

    def getCoordinateFields(self) -> list[str]:
        """Return the swarm fields used as coordinates on this cell DM.

        Not collective.

        See Also
        --------
        petsc.DMSwarmCellDMGetCoordinateFields

        """
        cdef PetscInt i = 0, ncf = 0
        cdef const char **cf = NULL
        CHKERR(DMSwarmCellDMGetCoordinateFields(self.cdm, &ncf, &cf))
        cdef list fields = []
        for i from 0 <= i < ncf:
            fields.append(bytes2str(cf[i]))
        return fields

# --------------------------------------------------------------------

del DMSwarmType
del DMSwarmMigrateType
del DMSwarmCollectType
del DMSwarmPICLayoutType



from petsc4py import PETSc
import unittest

class TestMatSchur(unittest.TestCase):
    def test(self):
        COMM = PETSc.COMM_WORLD
        TYPE = PETSc.Mat.Type.AIJ
        comm_size = COMM.getSize()
        A00 = PETSc.Mat().create(comm=COMM)
        A00.setType(TYPE)
        A00.setSizes([[2, 2*comm_size], [2, 2*comm_size]])
        A01 = PETSc.Mat().create(comm=COMM)
        A01.setType(TYPE)
        A01.setSizes([[2, 2*comm_size], [3, 3*comm_size]])
        A10 = PETSc.Mat().create(comm=COMM)
        A10.setType(TYPE)
        A10.setSizes([[3, 3*comm_size], [2, 2*comm_size]])
        A11 = PETSc.Mat().create(comm=COMM)
        A11.setType(TYPE)
        A11.setSizes([[3, 3*comm_size], [3, 3*comm_size]])
        S = PETSc.Mat().createSchurComplement(A00, A00, A01, A10, A11)
        M, N = S.getSize()
        self.assertEqual(M, 3*comm_size)
        self.assertEqual(N, 3*comm_size)
        m, n = S.getLocalSize()
        self.assertEqual(m, 3)
        self.assertEqual(n, 3)
        A00_dup, A00p_dup, A01_dup, A10_dup, A11_dup = S.getSchurComplementSubMatrices()
        self.assertEqual(A00_dup.id, A00.id)
        self.assertEqual(A00p_dup.id, A00.id)
        self.assertEqual(A01_dup.id, A01.id)
        self.assertEqual(A10_dup.id, A10.id)
        self.assertEqual(A11_dup.id, A11.id)

if __name__ == '__main__':
    unittest.main()



cdef extern from * nogil:

    ctypedef const char* PetscPCType "PCType"
    PetscPCType PCNONE
    PetscPCType PCJACOBI
    PetscPCType PCSOR
    PetscPCType PCLU
    PetscPCType PCQR
    PetscPCType PCSHELL
    PetscPCType PCBJACOBI
    PetscPCType PCMG
    PetscPCType PCEISENSTAT
    PetscPCType PCILU
    PetscPCType PCICC
    PetscPCType PCASM
    PetscPCType PCGASM
    PetscPCType PCKSP
    PetscPCType PCCOMPOSITE
    PetscPCType PCREDUNDANT
    PetscPCType PCSPAI
    PetscPCType PCNN
    PetscPCType PCCHOLESKY
    PetscPCType PCPBJACOBI
    PetscPCType PCVPBJACOBI
    PetscPCType PCMAT
    PetscPCType PCHYPRE
    PetscPCType PCPARMS
    PetscPCType PCFIELDSPLIT
    PetscPCType PCTFS
    PetscPCType PCML
    PetscPCType PCGALERKIN
    PetscPCType PCEXOTIC
    PetscPCType PCCP
    PetscPCType PCLSC
    PetscPCType PCPYTHON
    PetscPCType PCPFMG
    PetscPCType PCSYSPFMG
    PetscPCType PCREDISTRIBUTE
    PetscPCType PCSVD
    PetscPCType PCGAMG
    PetscPCType PCCHOWILUVIENNACL
    PetscPCType PCROWSCALINGVIENNACL
    PetscPCType PCSAVIENNACL
    PetscPCType PCBDDC
    PetscPCType PCKACZMARZ
    PetscPCType PCTELESCOPE
    PetscPCType PCPATCH
    PetscPCType PCLMVM
    PetscPCType PCHMG
    PetscPCType PCDEFLATION
    PetscPCType PCHPDDM
    PetscPCType PCH2OPUS

    ctypedef enum PetscPCSide "PCSide":
        PC_SIDE_DEFAULT
        PC_LEFT
        PC_RIGHT
        PC_SYMMETRIC

    ctypedef enum PetscPCASMType "PCASMType":
        PC_ASM_BASIC
        PC_ASM_RESTRICT
        PC_ASM_INTERPOLATE
        PC_ASM_NONE

    ctypedef enum PetscPCGASMType "PCGASMType":
        PC_GASM_BASIC
        PC_GASM_RESTRICT
        PC_GASM_INTERPOLATE
        PC_GASM_NONE

    ctypedef enum PetscPCMGType "PCMGType":
        PC_MG_MULTIPLICATIVE
        PC_MG_ADDITIVE
        PC_MG_FULL
        PC_MG_KASKADE

    ctypedef enum PetscPCMGCycleType "PCMGCycleType":
        PC_MG_CYCLE_V
        PC_MG_CYCLE_W

    ctypedef const char* PetscPCGAMGType "PCGAMGType"
    PetscPCGAMGType PCGAMGAGG
    PetscPCGAMGType PCGAMGGEO
    PetscPCGAMGType PCGAMGCLASSICAL

    ctypedef const char* PetscPCHYPREType "const char*"

    ctypedef enum PetscPCCompositeType "PCCompositeType":
        PC_COMPOSITE_ADDITIVE
        PC_COMPOSITE_MULTIPLICATIVE
        PC_COMPOSITE_SYMMETRIC_MULTIPLICATIVE
        PC_COMPOSITE_SPECIAL
        PC_COMPOSITE_SCHUR

    ctypedef enum PetscPCFieldSplitSchurPreType "PCFieldSplitSchurPreType":
        PC_FIELDSPLIT_SCHUR_PRE_SELF
        PC_FIELDSPLIT_SCHUR_PRE_SELFP
        PC_FIELDSPLIT_SCHUR_PRE_A11
        PC_FIELDSPLIT_SCHUR_PRE_USER
        PC_FIELDSPLIT_SCHUR_PRE_FULL

    ctypedef enum PetscPCFieldSplitSchurFactType "PCFieldSplitSchurFactType":
        PC_FIELDSPLIT_SCHUR_FACT_DIAG
        PC_FIELDSPLIT_SCHUR_FACT_LOWER
        PC_FIELDSPLIT_SCHUR_FACT_UPPER
        PC_FIELDSPLIT_SCHUR_FACT_FULL

    ctypedef enum PetscPCPatchConstructType "PCPatchConstructType":
        PC_PATCH_STAR
        PC_PATCH_VANKA
        PC_PATCH_PARDECOMP
        PC_PATCH_USER
        PC_PATCH_PYTHON

    ctypedef enum PetscPCHPDDMCoarseCorrectionType "PCHPDDMCoarseCorrectionType":
        PC_HPDDM_COARSE_CORRECTION_DEFLATED
        PC_HPDDM_COARSE_CORRECTION_ADDITIVE
        PC_HPDDM_COARSE_CORRECTION_BALANCED
        PC_HPDDM_COARSE_CORRECTION_NONE

    ctypedef enum PetscPCDeflationSpaceType "PCDeflationSpaceType":
        PC_DEFLATION_SPACE_HAAR
        PC_DEFLATION_SPACE_DB2
        PC_DEFLATION_SPACE_DB4
        PC_DEFLATION_SPACE_DB8
        PC_DEFLATION_SPACE_DB16
        PC_DEFLATION_SPACE_BIORTH22
        PC_DEFLATION_SPACE_MEYER
        PC_DEFLATION_SPACE_AGGREGATION
        PC_DEFLATION_SPACE_USER

    ctypedef enum PetscPCFailedReason "PCFailedReason":
        PC_SETUP_ERROR
        PC_NOERROR
        PC_FACTOR_STRUCT_ZEROPIVOT
        PC_FACTOR_NUMERIC_ZEROPIVOT
        PC_FACTOR_OUTMEMORY
        PC_FACTOR_OTHER
        PC_SUBPC_ERROR

    PetscErrorCode PCCreate(MPI_Comm, PetscPC*)
    PetscErrorCode PCDestroy(PetscPC*)
    PetscErrorCode PCView(PetscPC, PetscViewer)

    PetscErrorCode PCSetType(PetscPC, PetscPCType)
    PetscErrorCode PCGetType(PetscPC, PetscPCType*)

    PetscErrorCode PCSetOptionsPrefix(PetscPC, char[])
    PetscErrorCode PCAppendOptionsPrefix(PetscPC, char[])
    PetscErrorCode PCGetOptionsPrefix(PetscPC, char*[])
    PetscErrorCode PCSetFromOptions(PetscPC)

    PetscErrorCode PCSetFailedReason(PetscPC, PetscPCFailedReason)
    PetscErrorCode PCGetFailedReason(PetscPC, PetscPCFailedReason*)

    PetscErrorCode PCSetUp(PetscPC)
    PetscErrorCode PCReset(PetscPC)
    PetscErrorCode PCSetUpOnBlocks(PetscPC)

    PetscErrorCode PCApply(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCMatApply(PetscPC, PetscMat, PetscMat)
    PetscErrorCode PCApplyTranspose(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCMatApplyTranspose(PetscPC, PetscMat, PetscMat)
    PetscErrorCode PCApplySymmetricLeft(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCApplySymmetricRight(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCApplyRichardson(PetscPC, PetscVec, PetscVec, PetscVec, PetscReal, PetscReal, PetscReal, PetscInt)
    PetscErrorCode PCApplyBAorAB(PetscPC, PetscPCSide, PetscVec, PetscVec, PetscVec)
    PetscErrorCode PCApplyBAorABTranspose(PetscPC, PetscPCSide, PetscVec, PetscVec, PetscVec)

    PetscErrorCode PCGetDM(PetscPC, PetscDM*)
    PetscErrorCode PCSetDM(PetscPC, PetscDM)

    PetscErrorCode PCSetOperators(PetscPC, PetscMat, PetscMat)
    PetscErrorCode PCGetOperators(PetscPC, PetscMat*, PetscMat*)
    PetscErrorCode PCGetOperatorsSet(PetscPC, PetscBool*, PetscBool*)
    PetscErrorCode PCSetCoordinates(PetscPC, PetscInt, PetscInt, PetscReal[])
    PetscErrorCode PCSetUseAmat(PetscPC, PetscBool)
    PetscErrorCode PCGetUseAmat(PetscPC, PetscBool*)

    PetscErrorCode PCComputeExplicitOperator(PetscPC, PetscMat*)

    PetscErrorCode PCDiagonalScale(PetscPC, PetscBool*)
    PetscErrorCode PCDiagonalScaleLeft(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCDiagonalScaleRight(PetscPC, PetscVec, PetscVec)
    PetscErrorCode PCDiagonalScaleSet(PetscPC, PetscVec)

    PetscErrorCode PCBJacobiGetSubKSP(PetscPC, PetscInt*, PetscInt*, PetscKSP*[])

    PetscErrorCode PCASMSetType(PetscPC, PetscPCASMType)
    PetscErrorCode PCASMSetOverlap(PetscPC, PetscInt)
    PetscErrorCode PCASMSetLocalSubdomains(PetscPC, PetscInt, PetscIS[], PetscIS[])
    PetscErrorCode PCASMSetTotalSubdomains(PetscPC, PetscInt, PetscIS[], PetscIS[])
    PetscErrorCode PCASMGetSubKSP(PetscPC, PetscInt*, PetscInt*, PetscKSP*[])
    PetscErrorCode PCASMSetSortIndices(PetscPC, PetscBool)

    PetscErrorCode PCGASMSetType(PetscPC, PetscPCGASMType)
    PetscErrorCode PCGASMSetOverlap(PetscPC, PetscInt)

    PetscErrorCode PCGAMGSetType(PetscPC, PetscPCGAMGType)
    PetscErrorCode PCGAMGSetNlevels(PetscPC, PetscInt)
    PetscErrorCode PCGAMGSetNSmooths(PetscPC, PetscInt)

    PetscErrorCode PCHYPREGetType(PetscPC, PetscPCHYPREType*)
    PetscErrorCode PCHYPRESetType(PetscPC, PetscPCHYPREType)
    PetscErrorCode PCHYPRESetDiscreteCurl(PetscPC, PetscMat)
    PetscErrorCode PCHYPRESetDiscreteGradient(PetscPC, PetscMat)
    PetscErrorCode PCHYPRESetAlphaPoissonMatrix(PetscPC, PetscMat)
    PetscErrorCode PCHYPRESetBetaPoissonMatrix(PetscPC, PetscMat)
    PetscErrorCode PCHYPRESetEdgeConstantVectors(PetscPC, PetscVec, PetscVec, PetscVec)
    PetscErrorCode PCHYPRESetInterpolations(PetscPC, PetscInt, PetscMat, PetscMat[], PetscMat, PetscMat[])
    PetscErrorCode PCHYPREAMSSetInteriorNodes(PetscPC, PetscVec)

    PetscErrorCode PCFactorGetMatrix(PetscPC, PetscMat*)
    PetscErrorCode PCFactorSetZeroPivot(PetscPC, PetscReal)
    PetscErrorCode PCFactorSetShiftType(PetscPC, PetscMatFactorShiftType)
    PetscErrorCode PCFactorSetShiftAmount(PetscPC, PetscReal)
    PetscErrorCode PCFactorSetMatSolverType(PetscPC, PetscMatSolverType)
    PetscErrorCode PCFactorGetMatSolverType(PetscPC, PetscMatSolverType*)
    PetscErrorCode PCFactorSetUpMatSolverType(PetscPC)
    PetscErrorCode PCFactorSetFill(PetscPC, PetscReal)
    PetscErrorCode PCFactorSetColumnPivot(PetscPC, PetscReal)
    PetscErrorCode PCFactorReorderForNonzeroDiagonal(PetscPC, PetscReal)
    PetscErrorCode PCFactorSetMatOrderingType(PetscPC, PetscMatOrderingType)
    PetscErrorCode PCFactorSetReuseOrdering(PetscPC, PetscBool)
    PetscErrorCode PCFactorSetReuseFill(PetscPC, PetscBool)
    PetscErrorCode PCFactorSetUseInPlace(PetscPC)
    PetscErrorCode PCFactorSetAllowDiagonalFill(PetscPC)
    PetscErrorCode PCFactorSetPivotInBlocks(PetscPC, PetscBool)
    PetscErrorCode PCFactorSetLevels(PetscPC, PetscInt)
    PetscErrorCode PCFactorSetDropTolerance(PetscPC, PetscReal, PetscReal, PetscInt)

    PetscErrorCode PCFieldSplitSetType(PetscPC, PetscPCCompositeType)
    PetscErrorCode PCFieldSplitSetBlockSize(PetscPC, PetscInt)
    PetscErrorCode PCFieldSplitSetFields(PetscPC, char[], PetscInt, PetscInt*, PetscInt*)
    PetscErrorCode PCFieldSplitSetIS(PetscPC, char[], PetscIS)
    PetscErrorCode PCFieldSplitGetSubKSP(PetscPC, PetscInt*, PetscKSP*[])
    PetscErrorCode PCFieldSplitGetIS(PetscPC, char[], PetscIS*)
    PetscErrorCode PCFieldSplitSchurGetSubKSP(PetscPC, PetscInt*, PetscKSP*[])
    PetscErrorCode PCFieldSplitSetSchurPre(PetscPC, PetscPCFieldSplitSchurPreType, PetscMat)
    PetscErrorCode PCFieldSplitSetSchurFactType(PetscPC, PetscPCFieldSplitSchurFactType)

    PetscErrorCode PCCompositeSetType(PetscPC, PetscPCCompositeType)
    PetscErrorCode PCCompositeGetPC(PetscPC, PetscInt, PetscPC*)
    PetscErrorCode PCCompositeAddPCType(PetscPC, PetscPCType)
    PetscErrorCode PCCompositeAddPC(PetscPC, PetscPC)

    PetscErrorCode PCKSPGetKSP(PetscPC, PetscKSP*)

    PetscErrorCode PCSetReusePreconditioner(PetscPC, PetscBool)

    # --- MG ---
    PetscErrorCode PCMGSetType(PetscPC, PetscPCMGType)
    PetscErrorCode PCMGGetType(PetscPC, PetscPCMGType*)
    PetscErrorCode PCMGSetInterpolation(PetscPC, PetscInt, PetscMat)
    PetscErrorCode PCMGGetInterpolation(PetscPC, PetscInt, PetscMat*)
    PetscErrorCode PCMGSetRestriction(PetscPC, PetscInt, PetscMat)
    PetscErrorCode PCMGGetRestriction(PetscPC, PetscInt, PetscMat*)
    PetscErrorCode PCMGSetRScale(PetscPC, PetscInt, PetscVec)
    PetscErrorCode PCMGGetRScale(PetscPC, PetscInt, PetscVec*)
    PetscErrorCode PCMGGetSmoother(PetscPC, PetscInt, PetscKSP*)
    PetscErrorCode PCMGGetSmootherUp(PetscPC, PetscInt, PetscKSP*)
    PetscErrorCode PCMGGetSmootherDown(PetscPC, PetscInt, PetscKSP*)
    PetscErrorCode PCMGGetCoarseSolve(PetscPC, PetscKSP*)
    PetscErrorCode PCMGSetRhs(PetscPC, PetscInt, PetscVec)
    PetscErrorCode PCMGSetX(PetscPC, PetscInt, PetscVec)
    PetscErrorCode PCMGSetR(PetscPC, PetscInt, PetscVec)
    PetscErrorCode PCMGSetLevels(PetscPC, PetscInt, MPI_Comm*)
    PetscErrorCode PCMGGetLevels(PetscPC, PetscInt*)
    PetscErrorCode PCMGSetCycleType(PetscPC, PetscPCMGCycleType)
    PetscErrorCode PCMGSetCycleTypeOnLevel(PetscPC, PetscInt, PetscPCMGCycleType)

    # --- BDDC ---
    PetscErrorCode PCBDDCSetDiscreteGradient(PetscPC, PetscMat, PetscInt, PetscInt, PetscBool, PetscBool)
    PetscErrorCode PCBDDCSetDivergenceMat(PetscPC, PetscMat, PetscBool, PetscIS)
    PetscErrorCode PCBDDCSetChangeOfBasisMat(PetscPC, PetscMat, PetscBool)
    PetscErrorCode PCBDDCSetPrimalVerticesIS(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetPrimalVerticesLocalIS(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetCoarseningRatio(PetscPC, PetscInt)
    PetscErrorCode PCBDDCSetLevels(PetscPC, PetscInt)
    PetscErrorCode PCBDDCSetDirichletBoundaries(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetDirichletBoundariesLocal(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetNeumannBoundaries(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetNeumannBoundariesLocal(PetscPC, PetscIS)
    PetscErrorCode PCBDDCSetDofsSplitting(PetscPC, PetscInt, PetscIS[])
    PetscErrorCode PCBDDCSetDofsSplittingLocal(PetscPC, PetscInt, PetscIS[])
    PetscErrorCode PCBDDCSetLocalAdjacencyGraph(PetscPC, PetscInt, const PetscInt*, const PetscInt*, PetscCopyMode)

    # --- Patch ---
    ctypedef PetscErrorCode (*PetscPCPatchComputeOperator)(PetscPC,
                                                           PetscInt,
                                                           PetscVec,
                                                           PetscMat,
                                                           PetscIS,
                                                           PetscInt,
                                                           const PetscInt*,
                                                           const PetscInt*,
                                                           void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscPCPatchComputeFunction)(PetscPC,
                                                           PetscInt,
                                                           PetscVec,
                                                           PetscVec,
                                                           PetscIS,
                                                           PetscInt,
                                                           const PetscInt*,
                                                           const PetscInt*,
                                                           void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*PetscPCPatchConstructOperator)(PetscPC,
                                                             PetscInt*,
                                                             PetscIS**,
                                                             PetscIS*,
                                                             void*) except PETSC_ERR_PYTHON
    PetscErrorCode PCPatchGetSubKSP(PetscPC, PetscInt*, PetscKSP*[])
    PetscErrorCode PCPatchSetCellNumbering(PetscPC, PetscSection)
    PetscErrorCode PCPatchSetDiscretisationInfo(PetscPC, PetscInt, PetscDM*, PetscInt*, PetscInt*, const PetscInt**, const PetscInt*, PetscInt, const PetscInt*, PetscInt, const PetscInt*)
    PetscErrorCode PCPatchSetComputeOperator(PetscPC, PetscPCPatchComputeOperator, void*)
    PetscErrorCode PCPatchSetComputeOperatorInteriorFacets(PetscPC, PetscPCPatchComputeOperator, void*)
    PetscErrorCode PCPatchSetComputeFunction(PetscPC, PetscPCPatchComputeFunction, void*)
    PetscErrorCode PCPatchSetComputeFunctionInteriorFacets(PetscPC, PetscPCPatchComputeFunction, void*)
    PetscErrorCode PCPatchSetConstructType(PetscPC, PetscPCPatchConstructType, PetscPCPatchConstructOperator, void*)

    # --- HPDDM ---
    ctypedef PetscErrorCode (*PetscPCHPDDMAuxiliaryMat)(PetscMat,
                                                        PetscReal,
                                                        PetscVec,
                                                        PetscVec,
                                                        PetscReal,
                                                        PetscIS,
                                                        void*) except PETSC_ERR_PYTHON
    PetscErrorCode PCHPDDMSetAuxiliaryMat(PetscPC, PetscIS, PetscMat, PetscPCHPDDMAuxiliaryMat, void*)
    PetscErrorCode PCHPDDMSetRHSMat(PetscPC, PetscMat)
    PetscErrorCode PCHPDDMGetComplexities(PetscPC, PetscReal*, PetscReal*)
    PetscErrorCode PCHPDDMHasNeumannMat(PetscPC, PetscBool)
    PetscErrorCode PCHPDDMSetCoarseCorrectionType(PetscPC, PetscPCHPDDMCoarseCorrectionType)
    PetscErrorCode PCHPDDMGetCoarseCorrectionType(PetscPC, PetscPCHPDDMCoarseCorrectionType*)
    PetscErrorCode PCHPDDMGetSTShareSubKSP(PetscPC, PetscBool*)
    PetscErrorCode PCHPDDMSetDeflationMat(PetscPC, PetscIS, PetscMat)

    # --- SPAI ---
    PetscErrorCode PCSPAISetEpsilon(PetscPC, PetscReal)
    PetscErrorCode PCSPAISetNBSteps(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetMax(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetMaxNew(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetBlockSize(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetCacheSize(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetVerbose(PetscPC, PetscInt)
    PetscErrorCode PCSPAISetSp(PetscPC, PetscInt)

    # --- DEFLATION ---
    PetscErrorCode PCDeflationSetInitOnly(PetscPC, PetscBool)
    PetscErrorCode PCDeflationSetLevels(PetscPC, PetscInt)
    PetscErrorCode PCDeflationSetReductionFactor(PetscPC, PetscInt)
    PetscErrorCode PCDeflationSetCorrectionFactor(PetscPC, PetscScalar)
    PetscErrorCode PCDeflationSetSpaceToCompute(PetscPC, PetscPCDeflationSpaceType, PetscInt)
    PetscErrorCode PCDeflationSetSpace(PetscPC, PetscMat, PetscBool)
    PetscErrorCode PCDeflationSetProjectionNullSpaceMat(PetscPC, PetscMat)
    PetscErrorCode PCDeflationSetCoarseMat(PetscPC, PetscMat)
    PetscErrorCode PCDeflationGetCoarseKSP(PetscPC, PetscKSP*)
    PetscErrorCode PCDeflationGetPC(PetscPC, PetscPC*)

    # --- PYTHON ---
    PetscErrorCode PCPythonSetType(PetscPC, char[])
    PetscErrorCode PCPythonGetType(PetscPC, char*[])

# --------------------------------------------------------------------

cdef inline PC ref_PC(PetscPC pc):
    cdef PC ob = <PC> PC()
    ob.pc = pc
    CHKERR(PetscINCREF(ob.obj))
    return ob

cdef PetscErrorCode PCPatch_ComputeOperator(
    PetscPC pc,
    PetscInt point,
    PetscVec vec,
    PetscMat mat,
    PetscIS cells,
    PetscInt ndof,
    const PetscInt *dofmap,
    const PetscInt *dofmapWithAll,
    void *ctx) except PETSC_ERR_PYTHON with gil:
    cdef Vec Vec = ref_Vec(vec)
    cdef Mat Mat = ref_Mat(mat)
    cdef PC Pc = ref_PC(pc)
    cdef IS Is = ref_IS(cells)
    cdef object context = Pc.get_attr("__patch_compute_operator__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple
    (op, args, kargs) = context
    cdef PetscInt[:] pydofs = <PetscInt[:ndof]>dofmap
    cdef PetscInt[:] pydofsWithAll
    if dofmapWithAll != NULL:
        pydofsWithAll = <PetscInt[:ndof]>dofmapWithAll
        dofsall = asarray(pydofsWithAll)
    else:
        dofsall = None
    op(Pc, toInt(point), Vec, Mat, Is, asarray(pydofs), dofsall, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode PCPatch_ComputeFunction(
    PetscPC pc,
    PetscInt point,
    PetscVec vec,
    PetscVec out,
    PetscIS cells,
    PetscInt ndof,
    const PetscInt *dofmap,
    const PetscInt *dofmapWithAll,
    void *ctx) except PETSC_ERR_PYTHON with gil:
    cdef Vec Out = ref_Vec(out)
    cdef Vec Vec = ref_Vec(vec)
    cdef PC Pc = ref_PC(pc)
    cdef IS Is = ref_IS(cells)
    cdef object context = Pc.get_attr("__patch_compute_function__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple
    (op, args, kargs) = context
    cdef PetscInt[:] pydofs = <PetscInt[:ndof]>dofmap
    cdef PetscInt[:] pydofsWithAll = <PetscInt[:ndof]>dofmapWithAll
    op(Pc, toInt(point), Vec, Out, Is, asarray(pydofs), asarray(pydofsWithAll), *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode PCPatch_ComputeOperatorInteriorFacets(
    PetscPC pc,
    PetscInt point,
    PetscVec vec,
    PetscMat mat,
    PetscIS facets,
    PetscInt ndof,
    const PetscInt *dofmap,
    const PetscInt *dofmapWithAll,
    void *ctx) except PETSC_ERR_PYTHON with gil:
    cdef Vec Vec = ref_Vec(vec)
    cdef Mat Mat = ref_Mat(mat)
    cdef PC Pc = ref_PC(pc)
    cdef IS Is = ref_IS(facets)
    cdef object context = Pc.get_attr("__patch_compute_operator_interior_facets__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple
    (op, args, kargs) = context
    cdef PetscInt[:] pydofs = <PetscInt[:ndof]>dofmap
    cdef PetscInt[:] pydofsWithAll
    if dofmapWithAll != NULL:
        pydofsWithAll = <PetscInt[:ndof]>dofmapWithAll
        dofsall = asarray(pydofsWithAll)
    else:
        dofsall = None
    op(Pc, toInt(point), Vec, Mat, Is, asarray(pydofs), dofsall, *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode PCPatch_ComputeFunctionInteriorFacets(
    PetscPC pc,
    PetscInt point,
    PetscVec vec,
    PetscVec out,
    PetscIS facets,
    PetscInt ndof,
    const PetscInt *dofmap,
    const PetscInt *dofmapWithAll,
    void *ctx) except PETSC_ERR_PYTHON with gil:
    cdef Vec Out = ref_Vec(out)
    cdef Vec Vec = ref_Vec(vec)
    cdef PC Pc = ref_PC(pc)
    cdef IS Is = ref_IS(facets)
    cdef object context = Pc.get_attr("__patch_compute_function_interior_facets__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple
    (op, args, kargs) = context
    cdef PetscInt[:] pydofs = <PetscInt[:ndof]>dofmap
    cdef PetscInt[:] pydofsWithAll = <PetscInt[:ndof]>dofmapWithAll
    op(Pc, toInt(point), Vec, Out, Is, asarray(pydofs), asarray(pydofsWithAll), *args, **kargs)
    return PETSC_SUCCESS

cdef PetscErrorCode PCPatch_UserConstructOperator(
    PetscPC pc,
    PetscInt *n,
    PetscIS **userIS,
    PetscIS *userIterationSet,
    void *ctx) except PETSC_ERR_PYTHON with gil:
    cdef PC Pc = ref_PC(pc)
    cdef PetscInt i
    cdef object context = Pc.get_attr("__patch_construction_operator__")
    if context is None and ctx != NULL: context = <object>ctx
    assert context is not None and type(context) is tuple
    (op, args, kargs) = context
    (patches, iterationSet) = op(Pc, *args, **kargs)
    n[0] = <PetscInt>len(patches)
    CHKERR(PetscMalloc(<size_t>n[0]*sizeof(PetscIS), userIS))
    for i in range(n[0]):
        userIS[0][i] = (<IS?>patches[i]).iset
        CHKERR(PetscINCREF(<PetscObject*>&userIS[0][i]))
    userIterationSet[0] = (<IS?>iterationSet).iset
    CHKERR(PetscINCREF(<PetscObject*>&userIterationSet[0]))
    return PETSC_SUCCESS



# The user-defined Python class implementing the Jacobi method.
class myJacobi:
    # Setup the internal data. In this case, we access the matrix diagonal.
    def setUp(self, pc):
        _, P = pc.getOperators()
        self.D = P.getDiagonal()

    # Apply the preconditioner
    def apply(self, pc, x, y):
        y.pointwiseDivide(x, self.D)



# --------------------------------------------------------------------

cdef object functools = None
import functools

# --------------------------------------------------------------------

cdef class Log:
    """Logging support."""

    @classmethod
    def Stage(cls, name: str) -> LogStage:
        """Create a log stage.

        Not collective.

        Parameters
        ----------
        name
            Stage name.

        Returns
        -------
        stage : LogStage
            The log stage. If a stage already exists with name ``name`` then
            it is reused.

        See Also
        --------
        petsc.PetscLogStageRegister

        """
        if not name: raise ValueError("empty name")
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscLogStage stageid = -1
        cdef LogStage stage = get_LogStage(name)
        if stage is not None: return stage
        CHKERR(PetscLogStageFindId(cname, &stageid))
        if stageid == -1:
            CHKERR(PetscLogStageRegister(cname, &stageid))
        stage = reg_LogStage(name, stageid)
        return stage

    @classmethod
    def Class(cls, name: str) -> LogClass:
        """Create a log class.

        Not collective.

        Parameters
        ----------
        name
            Class name.

        Returns
        -------
        klass : LogClass
            The log class. If a class already exists with name ``name`` then
            it is reused.

        See Also
        --------
        petsc.PetscClassIdRegister

        """
        if not name: raise ValueError("empty name")
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscLogClass classid = -1
        cdef LogClass klass = get_LogClass(name)
        if klass is not None: return klass
        CHKERR(PetscLogClassFindId(cname, &classid))
        if classid == -1:
            CHKERR(PetscLogClassRegister(cname, &classid))
        klass = reg_LogClass(name, classid)
        return klass

    @classmethod
    def Event(cls, name: str, klass: LogClass | None = None) -> LogEvent:
        """Create a log event.

        Not collective.

        Parameters
        ----------
        name
            Event name.
        klass
            Log class. If `None`, defaults to ``PETSC_OBJECT_CLASSID``.

        Returns
        -------
        event : LogEvent
            The log event. If an event already exists with name ``name`` then
            it is reused.

        See Also
        --------
        petsc.PetscLogEventRegister

        """
        if not name: raise ValueError("empty name")
        cdef const char *cname = NULL
        name = str2bytes(name, &cname)
        cdef PetscLogClass classid = PETSC_OBJECT_CLASSID
        cdef PetscLogEvent eventid = -1
        if klass is not None: classid = klass
        cdef LogEvent event = get_LogEvent(name)
        if event is not None: return event
        CHKERR(PetscLogEventFindId(cname, &eventid))
        if eventid == -1:
            CHKERR(PetscLogEventRegister(cname, classid, &eventid))
        event = reg_LogEvent(name, eventid)
        return event

    @classmethod
    def begin(cls) -> None:
        """Turn on logging of objects and events.

        Collective.

        See Also
        --------
        petsc.PetscLogDefaultBegin

        """
        CHKERR(PetscLogDefaultBegin())

    @classmethod
    def view(cls, Viewer viewer=None) -> None:
        """Print the log.

        Collective.

        Parameters
        ----------
        viewer
            A `Viewer` instance or `None` for the default viewer.

        See Also
        --------
        petsc_options, petsc.PetscLogView

        """
        cdef PetscViewer vwr = NULL
        if viewer is not None: vwr = viewer.vwr
        if vwr == NULL: vwr = PETSC_VIEWER_STDOUT_WORLD
        CHKERR(PetscLogView(vwr))

    @classmethod
    def logFlops(cls, flops: float) -> None:
        """Add floating point operations to the current event.

        Not collective.

        Parameters
        ----------
        flops
            The number of flops to log.

        See Also
        --------
        petsc.PetscLogFlops

        """
        cdef PetscLogDouble cflops=flops
        CHKERR(PetscLogFlops(cflops))

    @classmethod
    def addFlops(cls, flops: float) -> None:
        """Add floating point operations to the current event.

        Not collective.

        Parameters
        ----------
        flops
            The number of flops to log.

        Notes
        -----
        This method exists for backward compatibility.

        See Also
        --------
        logFlops, petsc.PetscLogFlops

        """
        cdef PetscLogDouble cflops=flops
        CHKERR(PetscLogFlops(cflops))

    @classmethod
    def getFlops(cls) -> float:
        """Return the number of flops used on this processor since the program began.

        Not collective.

        Returns
        -------
        float
            Number of floating point operations.

        See Also
        --------
        petsc.PetscGetFlops

        """
        cdef PetscLogDouble cflops=0
        CHKERR(PetscGetFlops(&cflops))
        return cflops

    @classmethod
    def getTime(cls) -> float:
        """Return the current time of day in seconds.

        Collective.

        Returns
        -------
        wctime : float
            Current time.

        See Also
        --------
        petsc.PetscTime

        """
        cdef PetscLogDouble wctime=0
        CHKERR(PetscTime(&wctime))
        return wctime

    @classmethod
    def getCPUTime(cls) -> float:
        """Return the CPU time."""
        cdef PetscLogDouble cputime=0
        CHKERR(PetscGetCPUTime(&cputime))
        return cputime

    @classmethod
    def EventDecorator(cls, name=None, klass=None) -> Any:
        """Decorate a function with a `PETSc` event."""
        def decorator(func):
            @functools.wraps(func)
            def wrapped_func(*args, **kwargs):
                if name:
                    name_ = name
                else:
                    name_ = ".".join([func.__module__, getattr(func, "__qualname__", func.__name__)])
                with cls.Event(name_, klass):
                    return func(*args, **kwargs)
            return wrapped_func
        return decorator

    @classmethod
    def isActive(cls) -> bool:
        """Return whether logging is currently in progress.

        Not collective.

        See Also
        --------
        petsc.PetscLogIsActive

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(PetscLogIsActive(&flag))
        return toBool(flag)

# --------------------------------------------------------------------

cdef class LogStage:
    """Logging support for different stages."""

    cdef PetscLogStage id

    property id:
        """The log stage identifier."""
        def __get__(self) -> int:
            return self.id

    def __cinit__(self):
        self.id = 0

    def __int__(self):
        return <int> self.id

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, *exc):
        self.pop()

    #

    def push(self) -> None:
        """Push a stage on the logging stack.

        Logically collective.

        See Also
        --------
        LogStage.pop, petsc.PetscLogStagePush

        """
        CHKERR(PetscLogStagePush(self.id))

    def pop(self) -> None:
        """Pop a stage from the logging stack.

        Logically collective.

        See Also
        --------
        LogStage.push, petsc.PetscLogStagePop

        """
        <void>self # unused
        CHKERR(PetscLogStagePop())

    #

    def getName(self) -> str:
        """Return the current stage name."""
        cdef const char *cval = NULL
        CHKERR(PetscLogStageFindName(self.id, &cval))
        return bytes2str(cval)

    property name:
        """The current stage name."""
        def __get__(self) -> str:
            return self.getName()

        def __set__(self, value):
            <void>self; <void>value # unused
            raise TypeError("readonly attribute")

    #

    def activate(self) -> None:
        """Activate the stage.

        Logically collective.

        See Also
        --------
        petsc.PetscLogStageSetActive

        """
        CHKERR(PetscLogStageSetActive(self.id, PETSC_TRUE))

    def deactivate(self) -> None:
        """Deactivate the stage.

        Logically collective.

        See Also
        --------
        petsc.PetscLogStageSetActive

        """
        CHKERR(PetscLogStageSetActive(self.id, PETSC_FALSE))

    def getActive(self) -> bool:
        """Check if the stage is activated.

        Not collective.

        See Also
        --------
        petsc.PetscLogStageGetActive

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(PetscLogStageGetActive(self.id, &flag))
        return toBool(flag)

    def setActive(self, flag: bool) -> None:
        """Activate or deactivate the current stage.

        Logically collective.

        See Also
        --------
        petsc.PetscLogStageSetActive

        """
        cdef PetscBool tval = PETSC_FALSE
        if flag: tval = PETSC_TRUE
        CHKERR(PetscLogStageSetActive(self.id, tval))

    property active:
        """Whether the stage is activate."""
        def __get__(self) -> bool:
            return self.getActive()

        def __set__(self, value):
            self.setActive(value)

    #

    def getVisible(self) -> bool:
        """Return whether the stage is visible.

        Not collective.

        See Also
        --------
        LogStage.setVisible, petsc.PetscLogStageSetVisible

        """
        cdef PetscBool flag = PETSC_FALSE
        CHKERR(PetscLogStageGetVisible(self.id, &flag))
        return toBool(flag)

    def setVisible(self, flag: bool) -> None:
        """Set the visibility of the stage.

        Logically collective.

        Parameters
        ----------
        flag
            `True` to make the stage visible, `False` otherwise.

        See Also
        --------
        LogStage.getVisible, petsc.PetscLogStageSetVisible

        """
        cdef PetscBool tval = PETSC_FALSE
        if flag: tval = PETSC_TRUE
        CHKERR(PetscLogStageSetVisible(self.id, tval))

    property visible:
        """Whether the stage is visible."""
        def __get__(self) -> bool:
            return self.getVisible()

        def __set__(self, value):
            self.setVisible(value)


cdef dict stage_registry = {}

cdef LogStage get_LogStage(object name):
    return stage_registry.get(name)

cdef LogStage reg_LogStage(object name, PetscLogStage stageid):
    cdef LogStage stage = LogStage()
    stage.id = stageid
    stage_registry[name] = stage
    return stage

# --------------------------------------------------------------------

cdef class LogClass:
    """Logging support."""

    cdef PetscLogClass id

    property id:
        """The log class identifier."""
        def __get__(self) -> int:
            return self.id

    def __cinit__(self):
        self.id = PETSC_OBJECT_CLASSID

    def __int__(self):
        return <int> self.id

    #

    def getName(self) -> str:
        """Return the log class name."""
        cdef const char *cval = NULL
        CHKERR(PetscLogClassFindName(self.id, &cval))
        return bytes2str(cval)

    property name:
        """The log class name."""
        def __get__(self) -> str:
            return self.getName()

        def __set__(self, value):
            <void>self; <void>value # unused
            raise TypeError("readonly attribute")

    #

    def activate(self) -> None:
        """Activate the log class."""
        CHKERR(PetscLogClassActivate(self.id))

    def deactivate(self) -> None:
        """Deactivate the log class."""
        CHKERR(PetscLogClassDeactivate(self.id))

    def getActive(self) -> bool:
        """Not implemented."""
        <void>self # unused
        raise NotImplementedError

    def setActive(self, flag: bool) -> None:
        """Activate or deactivate the log class."""
        if flag:
            CHKERR(PetscLogClassActivate(self.id))
        else:
            CHKERR(PetscLogClassDeactivate(self.id))

    property active:
        """Log class activation."""
        def __get__(self) -> bool:
            return self.getActive()

        def __set__(self, value):
            self.setActive(value)


cdef dict class_registry = {}

cdef LogClass get_LogClass(object name):
    return class_registry.get(name)

cdef LogClass reg_LogClass(object name, PetscLogClass classid):
    cdef LogClass klass = LogClass()
    klass.id = classid
    class_registry[name] = klass
    return klass

# --------------------------------------------------------------------

cdef class LogEvent:
    """Logging support."""

    cdef PetscLogEvent id

    property id:
        """The log event identifier."""
        def __get__(self) -> int:
            return self.id

    def __cinit__(self):
        self.id = 0

    def __int__(self):
        return <int> self.id

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, *exc):
        self.end()

    def begin(self, *objs) -> None:
        """Log the beginning of a user event.

        Collective.

        Parameters
        ----------
        *objs
            objects associated with the event

        See Also
        --------
        petsc.PetscLogEventBegin

        """
        cdef PetscObject o[4]
        event_args2objs(objs, o)
        CHKERR(PetscLogEventBegin(self.id, o[0], o[1], o[2], o[3]))

    def end(self, *objs) -> None:
        """Log the end of a user event.

        Collective.

        Parameters
        ----------
        *objs
            Objects associated with the event.

        See Also
        --------
        petsc.PetscLogEventEnd

        """
        cdef PetscObject o[4]
        event_args2objs(objs, o)
        CHKERR(PetscLogEventEnd(self.id, o[0], o[1], o[2], o[3]))

    #
    def getName(self) -> str:
        """The current event name."""
        cdef const char *cval = NULL
        CHKERR(PetscLogEventFindName(self.id, &cval))
        return bytes2str(cval)

    property name:
        """The current event name."""
        def __get__(self) ->str:
            return self.getName()

        def __set__(self, value):
            <void>self; <void>value # unused
            raise TypeError("readonly attribute")

    #

    def activate(self) -> None:
        """Indicate that the event should be logged.

        Logically collective.

        See Also
        --------
        petsc.PetscLogEventActivate

        """
        CHKERR(PetscLogEventActivate(self.id))

    def deactivate(self) -> None:
        """Indicate that the event should not be logged.

        Logically collective.

        See Also
        --------
        petsc.PetscLogEventDeactivate

        """
        CHKERR(PetscLogEventDeactivate(self.id))

    def getActive(self) -> bool:
        """Not implemented."""
        <void>self # unused
        raise NotImplementedError

    def setActive(self, flag: bool) -> None:
        """Indicate whether or not the event should be logged.

        Logically collective.

        Parameters
        ----------
        flag
            Activate or deactivate the event.

        See Also
        --------
        petsc.PetscLogEventDeactivate, petsc.PetscLogEventActivate

        """
        if flag:
            CHKERR(PetscLogEventActivate(self.id))
        else:
            CHKERR(PetscLogEventDeactivate(self.id))

    property active:
        """Event activation."""
        def __get__(self) -> bool:
            return self.getActive()

        def __set__(self, value):
            self.setActive(value)

    def getActiveAll(self) -> bool:
        """Not implemented."""
        <void>self # unused
        raise NotImplementedError

    def setActiveAll(self, flag: bool) -> None:
        """Turn on logging of all events.

        Logically collective.

        Parameters
        ----------
        flag
            Activate (if `True`) or deactivate (if `False`) the logging of all events.

        See Also
        --------
        petsc.PetscLogEventSetActiveAll

        """
        cdef PetscBool tval = PETSC_FALSE
        if flag: tval = PETSC_TRUE
        CHKERR(PetscLogEventSetActiveAll(self.id, tval))

    property active_all:
        """All events activation."""
        def __get__(self) -> bool:
            self.getActiveAll()

        def __set__(self, value):
            self.setActiveAll(value)

    #

    def getPerfInfo(self, stage: int | None = None) -> dict:
        """Get the performance information about the given event in the given event.

        Not collective.

        Parameters
        ----------
        stage
            The stage number.

        Returns
        -------
        info : dict
            This structure is filled with the performance information.

        See Also
        --------
        petsc.PetscLogEventGetPerfInfo

        """
        cdef PetscEventPerfInfo info
        cdef PetscInt cstage = PETSC_DETERMINE
        if stage is not None: cstage = asInt(stage)
        CHKERR(PetscLogEventGetPerfInfo(<int>cstage, self.id, &info))
        return info

cdef dict event_registry = {}

cdef LogEvent get_LogEvent(object name):
    return event_registry.get(name)

cdef LogEvent reg_LogEvent(object name, PetscLogEvent eventid):
    cdef LogEvent event = LogEvent()
    event.id = eventid
    event_registry[name] = event
    return event

# --------------------------------------------------------------------
